/*
 * sbp2_1394.c - SBP-2 protocol driver for IEEE-1394
 *
 * Copyright (C) 2000 James Goodwin, Filanet Corporation (www.filanet.com)
 * jamesg@filanet.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * Brief Description:
 *
 * This driver implements the Serial Bus Protocol 2 (SBP-2) over IEEE-1394
 * under Linux. The SBP-2 driver is implemented as an IEEE-1394 high-level
 * driver. It also registers as a SCSI lower-level driver in order to accept
 * SCSI commands for transport using SBP-2.
 *
 * Driver Loading:
 *
 * Currently, the SBP-2 driver is supported only as a module. Because the 
 * Linux SCSI stack is not Plug-N-Play aware, module load order is 
 * important. Assuming the SCSI core drivers are either built into the 
 * kernel or already loaded as modules, you should load the IEEE-1394 modules 
 * in the following order:
 *
 * 	ieee1394 (e.g. insmod ieee1394)
 *	ohci1394 (e.g. insmod ohci1394)
 *	sbp2_1394 (e.g. insmod sbp2_1394)
 *
 * The SBP-2 driver will attempt to discover any attached SBP-2 devices when first
 * loaded, or after any IEEE-1394 bus reset (e.g. a hot-plug). It will then print 
 * out a debug message indicating if it was able to discover a SBP-2 device.
 *
 * Currently, the SBP-2 driver registers with the SCSI stack "before"
 * detecting SBP-2 devices on the IEEE-1394 bus. This means that we miss the
 * initial SCSI bus scan. In order to actually let the scsi stack "detect" a 
 * SBP-2 device, you must either use the procfs add-single-device, 
 * remove-single-device, or a shell script such as rescan-scsi-bus.sh.
 *
 * The easiest way to add/detect new SBP-2 devices is to run the shell script
 * rescan-scsi-bus.sh.
 *
 * As an alternative, you may manually add/remove SBP-2 devices via the procfs with
 * add-single-device <h> <b> <t> <l> or remove-single-device <h> <b> <t> <l>, where:
 *	<h> = host (starting at zero for first SCSI adapter)
 *	<b> = bus (normally zero)
 *	<t> = target (starting at zero for first SBP-2 device)
 *	<l> - lun (normally zero)
 *
 * e.g. To manually add/detect a new SBP-2 device
 *	echo "scsi add-single-device 0 0 0 0" > /proc/scsi/scsi
 *
 * e.g. To manually remove a SBP-2 device after it's been unplugged
 *	echo "scsi remove-single-device 0 0 0 0" > /proc/scsi/scsi
 *
 * e.g. To check to see which SBP-2/SCSI devices are currently registered
 * 	cat /proc/scsi/scsi
 *
 * After scanning for new SCSI devices (above), you may access any attached 
 * SBP-2 storage devices as if they were SCSI devices (e.g. mount /dev/sda1, 
 * fdisk, mkfs, etc.).
 *
 * Current Support:
 *
 * The SBP-2 driver is still in an early state, but supports a variety of devices.
 * I have read/written many gigabytes of data from/to SBP-2 drives, and have seen 
 * performance of more than 16 MBytes/s on individual drives (limit of the media 
 * transfer rate).
 *
 * Following are the devices that have been tested successfully:
 *
 *	- Western Digital (WD) IEEE-1394 hard drives
 *	- Maxtor IEEE-1394 hard drives
 *	- VST IEEE-1394 hard drives and Zip drives (several flavors)
 *	- LaCie IEEE-1394 hard drives (several flavors)
 *	- QPS IEEE-1394 CD-RW/DVD drives
 *	- Fujitsu IEEE-1394 MO drives
 *	- Sony IEEE-1394 CD-RW drives
 *	- Epson IEEE-1394 scanner
 *	- SBP-2 bridge-based devices (LSI and Oxford Semiconductor bridges)
 *
 * Current Issues:
 *
 *	- Hot-Plugging: Additional testing/debugging is needed to fully support
 *	  SBP-2 drive hot-plugging while the sbp2 module is loaded. Because the
 *	  rest of the SCSI stack does not handle I/O errors well, hot-plugging
 *	  of mounted devices is risky. 
 *
 *	- Error Handling: SCSI aborts and bus reset requests are handled somewhat
 *	  but the code needs additional debugging.
 *
 *	- Testing: Need to do additional hot-plug testing, stress testing, and 
 *	  device compatibility testing.
 *
 *	- IEEE-1394 Bus Management: There is currently little bus management
 *	  in the core IEEE-1394 stack. Because of this, the SBP-2 driver handles
 *	  detection of SBP-2 devices itself. This should be moved to the core
 *	  stack.
 *
 *	- I/O is currently serialized because of some stress issues with ohci1394.
 *	  Debugging in progress.
 *
 *	- The SBP-2 driver is currently only supported as a module. It would not take
 *	  much work to allow it to be compiled into the kernel, but you'd have to 
 *	  add some init code to the kernel to support this... and modules are much
 *	  more flexible anyway.   ;-)
 *
 * Core IEEE-1394 Stack Changes:
 *
 *	- The IEEE-1394 core stack guid code attempts to read the node unique id from
 *	  each attached device after a bus reset. It currently uses a block read
 *	  request to do this, which "upsets" certain not-well-behaved devices, such as
 *	  some CD-RW drives from QPS. If you have trouble with your IEEE-1394 storage 
 *	  device being detected after loading sbp2_1394, try commenting out the 
 *	  init_ieee1394_guid() lines at the bottom of ieee1394_core.c (and rebuild ieee1394.o). 
 *	  When an IEEE-1394 storage device is detected, you will see a message printed 
 *	  out by sbp2_1394.
 *
 *	- SBP-2 driver should now support the TILynx adapter, although I do not have one to 
 *	  test with. Am not sure what the max real payload supported by this adapter is,
 *	  but have heard it's somewhere around 512KB. You might need to either update
 *	  the max_rec field in pcilynx.h (misc. settings of bus info block in config rom)
 *	  from 0xf064a000 to something like 0xf0648000 (1KB payload) or 0xf0647000 (512 byte
 *	  payload). Alternatively, you can tune max speed and payload with #defines below. 
 *
 * History:
 *
 *	07/25/00 - Initial revision (JSG)
 *	08/11/00 - Following changes/bug fixes were made (JSG):
 *		   * Bug fix to SCSI procfs code (still needs to be synched with 2.4 kernel).
 *		   * Bug fix where request sense commands were actually sent on the bus.
 *		   * Changed bus reset/abort code to deal with devices that spin up quite
 *		     slowly (which result in SCSI time-outs).
 *		   * "More" properly pull information from device's config rom, for enumeration
 *		     of SBP-2 devices, and determining SBP-2 register offsets.
 *		   * Change Simplified Direct Access Device type to Direct Access Device type in
 *		     returned inquiry data, in order to make the SCSI stack happy.
 *		   * Modified driver to register with the SCSI stack "before" enumerating any attached
 *		     SBP-2 devices. This means that you'll have to use procfs scsi-add-device or 
 *		     some sort of script to discover new SBP-2 devices.
 *		   * Minor re-write of some code and other minor changes.
 *	08/28/00 - Following changes/bug fixes were made (JSG):
 *		   * Bug fixes to scatter/gather support (case of one s/g element)
 *		   * Updated direction table for scsi commands (mostly DVD commands)
 *		   * Retries when trying to detect SBP-2 devices (for slow devices)
 *		   * Slightly better error handling (previously none) when commands time-out.
 *		   * Misc. other bug fixes and code reorganization.
 *	09/13/00 - Following changes/bug fixes were made (JSG)
 *		   * Moved detection/enumeration code to a kernel thread which is woken up when IEEE-1394
 *		     bus resets occur.
 *		   * Added code to handle bus resets and hot-plugging while devices are mounted, but full
 *		     hot-plug support is not quite there yet.
 *		   * Now use speed map to determine speed and max payload sizes for ORBs
 *		   * Clean-up of code and reorganization 
 *	09/19/00 - Added better hot-plug support and other minor changes (JSG)
 *	10/15/00 - Fixes for latest 2.4.0 test kernel, minor fix for hot-plug race. (JSG)
 *	12/03/00 - Created pool of request packet structures for use in sending out sbp2 command
 *		   and agent reset requests. This removes the kmallocs/kfrees in the critical I/O paths,
 *		   and also deals with some subtle race conditions related to allocating and freeing
 *		   packets. (JSG)
 *      12/09/00 - Improved the sbp2 device detection by actually reading the root and unit 
 *		   directory (khk@khk.net)
 *	12/23/00 - Following changes/enhancements were made (JSG)
 *		   * Only do SCSI to RBC command conversion for Direct Access and Simplified
 *		     Direct Access Devices (this is pulled from the config rom root directory).
 *		     This is needed because doing the conversion for all device types broke the
 *		     Epson scanner. Still looking for a better way of determining when to convert
 *		     commands (for RBC devices). Thanks to khk for helping on this!
 *		   * Added ability to "emulate" physical dma support, for host adapters such as TILynx.
 *		   * Determine max payload and speed by also looking at the host adapter's max_rec field.
 */
 
 
/*
 * Change these defines if you have a bad IEEE-1394 controller that has trouble running
 * 2KB packets at 400mb. May also want to bump these down if testing with pcilynx.
 *
 * NOTE: On certain OHCI parts I have seen short packets on async transmit (probably 
 * due to PCI latency/throughput issues with the part). You can bump down the speed or max
 * payload if you are running into this problem (e.g. SPEED_S200, MAX_PAYLOAD_S200).
 */
 #define SBP2_MAX_SPEED_ALLOWED		SPEED_S400		/* 400mb */
 #define SBP2_MAX_PAYLOAD_ALLOWED	MAX_PAYLOAD_S400 	/* 2KB */
 
/*
 * Define this so that the sbp2 driver will do internal command retries on unit attention
 * check condition. A unit attention could happen if a bus reset occurs on 1394, and 
 * sometimes the upper scsi drivers do not properly deal with it (e.g. while file copies are 
 * happening). We work around this by clearing the unit attention and retrying the command
 * that failed.
 */
#define SBP2_DO_RETRY_ON_UNIT_ATTENTION

/*
 * Define this if you'd like the sbp2 driver to work with 1394 adapters which do not support
 * physical dma in hardware (such as the TILynx)... 
 */
 #define SBP2_EMULATE_PHYSICAL_DMA_SUPPORT
 

/*
 * Includes
 */

#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/string.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/blk.h>
#include <linux/smp_lock.h>
#include <asm/current.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/byteorder.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/scatterlist.h>

#include "ieee1394.h"
#include "ieee1394_types.h"
#include "ieee1394_core.h"
#include "hosts.h"
#include "highlevel.h"
#include "ieee1394_transactions.h"
#include "../scsi/scsi.h"
#include "../scsi/hosts.h"
#include "../scsi/sd.h"
#include "sbp2_1394.h"

/*
 * Debug spewing control. Define SBP2_DEBUG_SPEWING or SBP2_DEBUG_SPEWING_MUCHO
 * if you'd like to see lots of messages...
 */

/* #define SBP2_DEBUG_SPEWING */ 
/* #define SBP2_DEBUG_SPEWING_MUCHO */  

#ifdef SBP2_DEBUG_SPEWING_MUCHO
#define SBP2_DEBUG(fmt, args...)	HPSB_ERR(fmt, ## args)	
#define SBP2_INFO(fmt, args...)		HPSB_ERR(fmt, ## args)	
#define SBP2_NOTICE(fmt, args...)	HPSB_ERR(fmt, ## args)	
#define SBP2_WARN(fmt, args...)		HPSB_ERR(fmt, ## args)
#else
#ifdef SBP2_DEBUG_SPEWING
#define SBP2_DEBUG(fmt, args...)	HPSB_DEBUG(fmt, ## args)
#define SBP2_INFO(fmt, args...)		HPSB_INFO(fmt, ## args)
#define SBP2_NOTICE(fmt, args...)	HPSB_NOTICE(fmt, ## args)
#define SBP2_WARN(fmt, args...)		HPSB_WARN(fmt, ## args)
#else 
#define SBP2_DEBUG(fmt, args...)	
#define SBP2_INFO(fmt, args...)	
#define SBP2_NOTICE(fmt, args...)	
#define SBP2_WARN(fmt, args...)	
#endif
#endif
#define SBP2_ERR(fmt, args...)		HPSB_ERR(fmt, ## args)

/*
 * Globals
 */

Scsi_Host_Template *global_scsi_tpnt = NULL;

LIST_HEAD(sbp2_host_info_list);
static int sbp2_host_count = 0;
spinlock_t sbp2_host_info_lock = SPIN_LOCK_UNLOCKED;

static struct hpsb_highlevel *sbp2_hl_handle = NULL;

static struct hpsb_highlevel_ops sbp2_hl_ops = {
	sbp2_add_host,
	sbp2_remove_host,
	sbp2_host_reset,
	NULL,
	NULL
};

static struct hpsb_address_ops sbp2_ops = {
	write: sbp2_handle_status_write,
};

#ifdef SBP2_EMULATE_PHYSICAL_DMA_SUPPORT
static struct hpsb_address_ops sbp2_physdma_ops = {
	read: sbp2_handle_physdma_read,
	write: sbp2_handle_physdma_write,
};
#endif

/**************************************
 * General utility functions
 **************************************/

/*
 * Converts a buffer from be32 to cpu byte ordering. Length is in bytes.
 */
static __inline__ void sbp2util_be32_to_cpu_buffer(void *buffer, int length)
{
	u32 *temp = buffer;
	int i;

	for (i=0; i<(length >> 2); i++) {
		temp[i] = be32_to_cpu(temp[i]);
	}
	return;
}

/*
 * Converts a buffer from cpu to be32 byte ordering. Length is in bytes.
 */
static __inline__ void sbp2util_cpu_to_be32_buffer(void *buffer, int length)
{
	u32 *temp = buffer;
	int i;

	for (i=0; i<(length >> 2); i++) {
		temp[i] = cpu_to_be32(temp[i]);
	}
	return;
}

/*
 * This function does quadlet sized reads (used by detection code)
 */
static int sbp2util_read_quadlet(struct sbp2scsi_host_info *hi, nodeid_t node, u64 addr,
				 quadlet_t *buffer)
{
	int retval = 0;
	int retry_count = 3;

	/*
	 * Retry a couple times if needed (for slow devices)
	 */
	do {

		retval = hpsb_read(hi->host, node, addr, buffer, 4);

		if (retval) {
			SBP2_DEBUG("sbp2: sbp2util_read_quadlet data packet error");
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(HZ/50);	/* 20ms delay */
		}

		retry_count--;

	} while (retval && retry_count);

	return(retval);
}

/*
 * This function returns the address of the unit directory.
 */
static int sbp2util_unit_directory(struct sbp2scsi_host_info *hi, nodeid_t node_id, u64 *unit_directory_addr)
{
	quadlet_t root_directory_length, current_quadlet;
	u64 current_addr;
	int length, i;

	/*
	 * First, read the first quadlet of the root directory to determine its size
	 */
	if (sbp2util_read_quadlet(hi, LOCAL_BUS | node_id, CONFIG_ROM_ROOT_DIR_BASE, 
				  &root_directory_length)) {
		SBP2_DEBUG("sbp2: Error reading root directory length - bad status");
		return(-EIO);   
	}

	current_addr = CONFIG_ROM_ROOT_DIR_BASE;
	length = be32_to_cpu(root_directory_length) >> 16;

	/*
	 * Step through the root directory and look for the "Unit_Directory entry", which
	 * contains the offset to the unit directory.
	 */
	for (i=0; i < length; i++) {

		current_addr += 4;

		if (sbp2util_read_quadlet(hi, LOCAL_BUS | node_id, current_addr, &current_quadlet)) {
			SBP2_DEBUG("sbp2: Error reading at address 0x%08x%08x - bad status", 
				   (unsigned int) ((current_addr) >> 32), (unsigned int) ((current_addr) & 0xffffffff));
			return(-EIO);   
		}

		/*  
		 * Check for unit directory offset tag 
		 */
		if ((be32_to_cpu(current_quadlet) >> 24) == SBP2_UNIT_DIRECTORY_OFFSET_KEY) {
			*unit_directory_addr = current_addr + 4 * ((be32_to_cpu(current_quadlet) & 0xffffff));
			SBP2_DEBUG("sbp2: unit_directory_addr = %x", (unsigned int) unit_directory_addr);
		}
	}

	return(0);
}

/*
 * This function is called to initially create a packet pool for use in sbp2 I/O requests.
 * This packet pool is used when sending out sbp2 command and agent reset requests, and 
 * allows us to remove all kmallocs/kfrees from the critical I/O paths.
 */
static int sbp2util_create_request_packet_pool(struct sbp2scsi_host_info *hi)
{
	struct hpsb_packet *packet;
	int i;
	unsigned long flags;

	/*
	 * Create SBP2_MAX_REQUEST_PACKETS number of request packets.
	 */
	spin_lock_irqsave(&hi->sbp2_request_packet_lock, flags);
	for (i=0; i<SBP2_MAX_REQUEST_PACKETS; i++) {

		/*
		 * Max payload of 8 bytes since the sbp2 command request uses a payload of 
		 * 8 bytes, and agent reset is a quadlet write request. Bump this up if we
		 * plan on using this pool for other stuff.
		 */
		packet = alloc_hpsb_packet(8);

		if (!packet) {
			SBP2_ERR("sbp2: sbp2util_create_request_packet_pool - packet allocation failed!");
			spin_unlock_irqrestore(&hi->sbp2_request_packet_lock, flags);
			return(-ENOMEM);
		}

		/* 
		 * Put these request packets into a free list
		 */
		INIT_LIST_HEAD(&hi->request_packet[i].list);
		hi->request_packet[i].packet = packet;
		list_add_tail(&hi->request_packet[i].list, &hi->sbp2_req_free);

	}
	spin_unlock_irqrestore(&hi->sbp2_request_packet_lock, flags);

	return(0);
}

/*
 * This function is called to remove the packet pool. It is called when the sbp2 driver is unloaded.
 */
static void sbp2util_remove_request_packet_pool(struct sbp2scsi_host_info *hi)
{
	struct list_head *lh;
	struct sbp2_request_packet *request_packet;
	unsigned long flags;

	/* 
	 * Go through free list releasing packets
	 */
	spin_lock_irqsave(&hi->sbp2_request_packet_lock, flags);
	while (!list_empty(&hi->sbp2_req_free)) {

		lh = hi->sbp2_req_free.next;
		list_del(lh);

		request_packet = list_entry(lh, struct sbp2_request_packet, list);

		/*
		 * Free the hpsb packets that we allocated for the pool
		 */
		if (request_packet) {
			free_hpsb_packet(request_packet->packet);
		}

	}
	spin_unlock_irqrestore(&hi->sbp2_request_packet_lock, flags);

	return;
}

/*
 * This function is called to retrieve a block write packet from our packet pool. This function is
 * used in place of calling alloc_hpsb_packet (which costs us three kmallocs). Instead we 
 * just pull out a free request packet and re-initialize values in it. I'm sure this can still
 * stand some more optimization. 
 */
static struct sbp2_request_packet *sbp2util_allocate_write_request_packet(struct sbp2scsi_host_info *hi,
									  nodeid_t node, u64 addr,
									  size_t data_size,
									  quadlet_t data) {
	struct list_head *lh;
	struct sbp2_request_packet *request_packet = NULL;
	struct hpsb_packet *packet;
	unsigned long flags;

	spin_lock_irqsave(&hi->sbp2_request_packet_lock, flags);
	if (!list_empty(&hi->sbp2_req_free)) {

		/*
		 * Pull out a free request packet
		 */
		lh = hi->sbp2_req_free.next;
		list_del(lh);

		request_packet = list_entry(lh, struct sbp2_request_packet, list);
		packet = request_packet->packet;

		/*
		 * Initialize the packet (this is really initialization the core 1394 stack should do,
		 * but I'm doing it myself to avoid the overhead).
		 */
		packet->data_size = data_size;
		INIT_LIST_HEAD(&packet->list);
		sema_init(&packet->state_change, 0);
		packet->state = unused;
		packet->generation = get_hpsb_generation();
		packet->data_be = 1;

		packet->host = hi->host;
		packet->tlabel = get_tlabel(hi->host, node, 1);
		packet->node_id = node;

		if (!data_size) {
			fill_async_writequad(packet, addr, data);
		} else {
			fill_async_writeblock(packet, addr, data_size);         
		}

		/*
		 * Set up a task queue completion routine, which returns the packet to the free list
		 * and releases the tlabel
		 */
		request_packet->tq.next = NULL;
		request_packet->tq.sync = 0;
		request_packet->tq.routine = (void (*)(void*))sbp2util_free_request_packet;
		request_packet->tq.data = request_packet;
		request_packet->hi_context = hi;
		queue_task(&request_packet->tq, &packet->complete_tq);

		/*
		 * Now, put the packet on the in-use list
		 */
		list_add_tail(&request_packet->list, &hi->sbp2_req_inuse);

	} else {
		SBP2_ERR("sbp2: sbp2util_allocate_request_packet - no packets available!");
	}
	spin_unlock_irqrestore(&hi->sbp2_request_packet_lock, flags);

	return(request_packet);
}

/*
 * This function is called to return a packet to our packet pool. It is also called as a 
 * completion routine when a request packet is completed.
 */
static void sbp2util_free_request_packet(struct sbp2_request_packet *request_packet)
{
	unsigned long flags;
	struct sbp2scsi_host_info *hi = request_packet->hi_context;

	/*
	 * Free the tlabel, and return the packet to the free pool
	 */
	spin_lock_irqsave(&hi->sbp2_request_packet_lock, flags);
	free_tlabel(hi->host, LOCAL_BUS | request_packet->packet->node_id, request_packet->packet->tlabel);
	list_del(&request_packet->list);
	list_add_tail(&request_packet->list, &hi->sbp2_req_free);
	spin_unlock_irqrestore(&hi->sbp2_request_packet_lock, flags);

	return;
}


/*********************************************
 * IEEE-1394 core driver stack related section
 *********************************************/

/*
 * This function is called at SCSI init in order to register our driver with the
 * IEEE-1394 stack
 */
int sbp2_init(void)
{
	SBP2_DEBUG("sbp2: sbp2_init");

	/*
	 * Register our high level driver with 1394 stack
	 */
	sbp2_hl_handle = hpsb_register_highlevel(SBP2_DEVICE_NAME, &sbp2_hl_ops);

	if (sbp2_hl_handle == NULL) {
		SBP2_ERR("sbp2: sbp2 failed to register with ieee1394 highlevel");
		return(-ENOMEM);
	}

	/*
	 * Register our sbp2 status address space...
	 */
	hpsb_register_addrspace(sbp2_hl_handle, &sbp2_ops, SBP2_STATUS_FIFO_ADDRESS,
				SBP2_STATUS_FIFO_ADDRESS + sizeof(struct sbp2_status_block));

	/*
	 * Register physical dma address space... used for adapters not supporting hardware phys dma.
	 */
#ifdef SBP2_EMULATE_PHYSICAL_DMA_SUPPORT
	hpsb_register_addrspace(sbp2_hl_handle, &sbp2_physdma_ops, 0x0ULL, 0xfffffffcULL);
#endif

	return(0);
}

/*
 * This function is called from cleanup module, or during shut-down, in order to 
 * unregister our driver
 */
void sbp2_cleanup(void)
{
	SBP2_DEBUG("sbp2: sbp2_cleanup");

	if (sbp2_hl_handle) {
		hpsb_unregister_highlevel(sbp2_hl_handle);
		sbp2_hl_handle = NULL;
	}
	return;
}

/*
 * This function is called after registering our operations in sbp2_init. We go ahead and
 * allocate some memory for our host info structure, and init some structures.
 */
static void sbp2_add_host(struct hpsb_host *host)
{
	struct sbp2scsi_host_info *hi;

	SBP2_DEBUG("sbp2: sbp2_add_host");

	/*
	 * Allocate some memory for our host info structure
	 */
	hi = (struct sbp2scsi_host_info *)kmalloc(sizeof(struct sbp2scsi_host_info), GFP_KERNEL);

	if (hi != NULL) {

		/*
		 * Initialize some host stuff
		 */
		memset(hi, 0, sizeof(struct sbp2scsi_host_info));
		INIT_LIST_HEAD(&hi->list);
		INIT_LIST_HEAD(&hi->sbp2_req_inuse);
		INIT_LIST_HEAD(&hi->sbp2_req_free);
		hi->host = host;
		hi->sbp2_command_lock = SPIN_LOCK_UNLOCKED;
		hi->sbp2_request_packet_lock = SPIN_LOCK_UNLOCKED;

		/*
		 * Create our request packet pool (pool of packets for use in I/O)
		 */
		if (sbp2util_create_request_packet_pool(hi)) {
			SBP2_ERR("sbp2: sbp2util_create_request_packet_pool failed!");
			return;
		}

		spin_lock_irq(&sbp2_host_info_lock);
		list_add_tail(&hi->list, &sbp2_host_info_list);
		sbp2_host_count++;
		spin_unlock_irq(&sbp2_host_info_lock);

		/*
		 * Initialize us to bus reset in progress
		 */
		hi->bus_reset_in_progress = 1;

		/*
		 * Register our host with the SCSI stack. 
		 */
		sbp2scsi_register_scsi_host(hi);

		/*
		 * Start our kernel thread to deal with sbp2 device detection
		 */
		init_waitqueue_head(&hi->sbp2_detection_wait);
		hi->sbp2_detection_pid = 0;
		hi->sbp2_detection_pid = kernel_thread(sbp2_detection_thread, hi, CLONE_FS | CLONE_FILES | CLONE_SIGHAND);

	}

	return;
}

/*
 * This fuction returns a host info structure from the host structure, in case we have multiple hosts
 */
static struct sbp2scsi_host_info *sbp2_find_host_info(struct hpsb_host *host) {
	struct list_head *lh;
	struct sbp2scsi_host_info *hi;

	lh = sbp2_host_info_list.next;
	while (lh != &sbp2_host_info_list) {
		hi = list_entry(lh, struct sbp2scsi_host_info, list);
		if (hi->host == host) {
			return hi;
		}
		lh = lh->next;
	}

	return(NULL);
}

/*
 * This function is called when the host is removed
 */
static void sbp2_remove_host(struct hpsb_host *host)
{
	struct sbp2scsi_host_info *hi;
	int i;

	SBP2_DEBUG("sbp2: sbp2_remove_host");

	spin_lock_irq(&sbp2_host_info_lock);
	hi = sbp2_find_host_info(host);

	if (hi != NULL) {

		/*
		 * Need to remove any attached SBP-2 devices. Also make sure to logout of all devices.
		 */
		for (i=0; i<SBP2SCSI_MAX_SCSI_IDS; i++) {
			if (hi->scsi_id[i]) {
				sbp2_logout_device(hi, hi->scsi_id[i]);
				hi->scsi_id[i]->validated = 0;
			}
		}

		sbp2_remove_unvalidated_devices(hi);

		list_del(&hi->list);
		sbp2_host_count--;
	}
	spin_unlock_irq(&sbp2_host_info_lock);

	if (hi == NULL) {
		SBP2_ERR("sbp2: attempt to remove unknown host 0x%p", host);
		return;
	}

	/*
	 * Remove the packet pool (release the packets)
	 */
	sbp2util_remove_request_packet_pool(hi);

	/* 
	 * Kill our detection thread 
	 */
	if (hi->sbp2_detection_pid >= 0) {
		kill_proc(hi->sbp2_detection_pid, SIGINT, 1);
	}

	/*
	 * Give the detection thread a little time to exit
	 */
	current->state = TASK_INTERRUPTIBLE;
	schedule_timeout(HZ);	/* 1 second delay */

	kfree(hi);
	hi = NULL;

	return;
}

/*
 * This is our sbp2 detection thread. It is signalled when bus resets occur
 * so that we can find and initialize any sbp2 devices. 
 */
static int sbp2_detection_thread(void *__hi)
{
	struct sbp2scsi_host_info *hi = (struct sbp2scsi_host_info *)__hi;

	SBP2_DEBUG("sbp2: sbp2_detection_thread");

	lock_kernel();

	/*
	 * This thread doesn't need any user-level access,
	 * so get rid of all our resources
	 */
	exit_mm(current);
	exit_files(current);
	exit_fs(current);

	/* 
	 * Set-up a nice name
	 */
	strcpy(current->comm, "sbp2_1394");

	while ((!signal_pending(current)) && hi) {

		/*
		 * Process our bus reset now
		 */
		if (hi) {
			MOD_INC_USE_COUNT;
			sbp2_bus_reset_handler(hi);
			MOD_DEC_USE_COUNT;
		}

		/*
		 * Sleep until next bus reset
		 */
		if (hi) {
			interruptible_sleep_on(&hi->sbp2_detection_wait);
		}
	}

	return(0);
}

/*
 * This function is where we first pull the node unique ids, and then allocate memory and register
 * a SBP-2 device
 */
static int sbp2_start_device(struct sbp2scsi_host_info *hi, int node_id)
{
	quadlet_t node_unique_id_lo, node_unique_id_hi;
	u64 node_unique_id;
	struct scsi_id_instance_data *scsi_id = NULL;
	int i;

	SBP2_DEBUG("sbp2: sbp2_start_device");

	/*
	 * Let's read the node unique id off of the device (using two quadlet reads for hi and lo)
	 */
	if (sbp2util_read_quadlet(hi, LOCAL_BUS | node_id, CONFIG_ROM_NODE_UNIQUE_ID_HI_ADDRESS, 
				  &node_unique_id_hi)) {
		SBP2_DEBUG("sbp2: Error reading node unique id - bad status");
		return(-EIO);
	}

	if (sbp2util_read_quadlet(hi, LOCAL_BUS | node_id, CONFIG_ROM_NODE_UNIQUE_ID_LO_ADDRESS, 
				  &node_unique_id_lo)) {
		SBP2_DEBUG("sbp2: Error reading node unique id - bad status");
		return(-EIO);
	}

	/*
	 * Spit out the node unique ids we got
	 */
	SBP2_DEBUG("sbp2: Node %x, node unique id hi = %x", (LOCAL_BUS | node_id), (unsigned int) node_unique_id_hi);
	SBP2_DEBUG("sbp2: Node %x, node unique id lo = %x", (LOCAL_BUS | node_id), (unsigned int) node_unique_id_lo);

	node_unique_id = (((u64)node_unique_id_hi) << 32) | ((u64)node_unique_id_lo);

	/*
	 * First, we need to find out whether this is a "new" SBP-2 device plugged in, or one that already
	 * exists and is initialized. We do this by looping through our scsi id instance data structures
	 * looking for matching node unique ids.
	 */
	for (i=0; i<SBP2SCSI_MAX_SCSI_IDS; i++) {

		if (hi->scsi_id[i]) {

			if (hi->scsi_id[i]->node_unique_id == node_unique_id) {

				/*
				 * Update our node id
				 */
				hi->scsi_id[i]->node_id = node_id;

				/*
				 * Mark the device as validated, since it still exists on the bus
				 */
				hi->scsi_id[i]->validated = 1;
				SBP2_DEBUG("sbp2: SBP-2 device re-validated, SCSI ID = %x", (unsigned int) i);

				/*
				 * Reconnect to the sbp-2 device
				 */
				if (sbp2_reconnect_device(hi, hi->scsi_id[i])) {

					/*
					 * Ok, reconnect has failed. Perhaps we didn't reconnect fast enough. Try
					 * doing a regular login.
					 */
					if (sbp2_login_device(hi, hi->scsi_id[i])) {

						/*
						 * Login failed too... so, just mark him as unvalidated, so that he gets cleaned up
						 * later
						 */
						SBP2_ERR("sbp2: sbp2_reconnect_device failed!");
						hi->scsi_id[i]->validated = 0;
					}
				}

				if (hi->scsi_id[i]->validated) {

					/*
					 * Set max retries to something large on the device
					 */
					sbp2_set_busy_timeout(hi, hi->scsi_id[i]);

					/*
					 * Enable unsolicited status on the device (not needed right now)
					 */
#ifdef SBP2_ENABLE_UNSOLICITED_STATUS
					sbp2_enable_unsolicited_status(hi, hi->scsi_id[i]);
#endif

					/*
					 * Do a SBP-2 fetch agent reset
					 */
					sbp2_agent_reset(hi, hi->scsi_id[i], 0);

					/*
					 * Get the max speed and packet size that we can use
					 */
					sbp2_max_speed_and_size(hi, hi->scsi_id[i]);

				}

				/*
				 * Nothing more to do, since we found the device
				 */
				return(0);

			}
		}
	}

	/*
	 * This really is a "new" device plugged in. Let's allocate memory for our scsi id instance data
	 */
	scsi_id = (struct scsi_id_instance_data *)kmalloc(sizeof(struct scsi_id_instance_data), GFP_KERNEL);

	if (!scsi_id) {
		SBP2_ERR("sbp2: Could not allocate scsi id instance data");
		return(-ENOMEM);
	}

	/*
	 * Initialize some of the fields in this structure
	 */
	memset(scsi_id, 0, sizeof(struct scsi_id_instance_data)); 
	scsi_id->node_id = node_id;
	scsi_id->node_unique_id = node_unique_id;
	scsi_id->validated = 1;
	scsi_id->speed_code = SPEED_S100;
	scsi_id->max_payload_size = MAX_PAYLOAD_S100;

	init_waitqueue_head(&scsi_id->sbp2_login_wait);

	/*
	 * Find an empty spot to stick our scsi id instance data. 
	 */
	for (i=0; i<SBP2SCSI_MAX_SCSI_IDS; i++) {
		if (!hi->scsi_id[i]) {
			hi->scsi_id[i] = scsi_id;
			SBP2_DEBUG("sbp2: New SBP-2 device inserted, SCSI ID = %x", (unsigned int) i);
			break;
		}
	}

	/*
	 * Make sure we are not out of space
	 */
	if (i >= SBP2SCSI_MAX_SCSI_IDS) {
		SBP2_ERR("sbp2: No slots left for SBP-2 device");
		return(-EBUSY);
	}

	/*
	 * Login to the sbp-2 device
	 */
	if (sbp2_login_device(hi, hi->scsi_id[i])) {

		/*
		 * Login failed... so, just mark him as unvalidated, so that he gets cleaned up later
		 */
		SBP2_ERR("sbp2: sbp2_login_device failed");
		hi->scsi_id[i]->validated = 0;
	}

	if (hi->scsi_id[i]->validated) {

		/*
		 * Set max retries to something large on the device
		 */
		sbp2_set_busy_timeout(hi, hi->scsi_id[i]);

		/*
		 * Enable unsolicited status on the device (not needed right now)
		 */
#ifdef SBP2_ENABLE_UNSOLICITED_STATUS
		sbp2_enable_unsolicited_status(hi, hi->scsi_id[i]);
#endif

		/*
		 * Do a SBP-2 fetch agent reset
		 */
		sbp2_agent_reset(hi, hi->scsi_id[i], 0);

		/*
		 * Get the max speed and packet size that we can use
		 */
		sbp2_max_speed_and_size(hi, hi->scsi_id[i]);

	}

	return(0);
}

/*
 * This function trys to determine if a device is a valid SBP-2 device
 */
static int sbp2_check_device(struct sbp2scsi_host_info *hi, int node_id)
{
	quadlet_t unit_spec_id_data = 0, unit_sw_ver_data = 0;
	quadlet_t unit_directory_length, current_quadlet;
	u64 unit_directory_addr, current_addr;
	unsigned int i, length;

	SBP2_DEBUG("sbp2: sbp2_check_device");

	/*
	 * Let's try and read the unit spec id and unit sw ver to determine if this is an SBP2 device...
	 */

	if (sbp2util_unit_directory(hi, LOCAL_BUS | node_id, &unit_directory_addr)) {
		SBP2_DEBUG("sbp2: Error reading unit directory address - bad status");
		return(-EIO);   
	}

	/*
	 * Read the size of the unit directory
	 */
	if (sbp2util_read_quadlet(hi, LOCAL_BUS | node_id, unit_directory_addr, 
				  &unit_directory_length)) {
		SBP2_DEBUG("sbp2: Error reading root directory length - bad status");
		return(-EIO);   
	}

	current_addr = unit_directory_addr;
	length = be32_to_cpu(unit_directory_length) >> 16;

	/*
	 * Now, step through the unit directory and look for the unit_spec_ID and the unit_sw_version
	 */
	for (i=0; i < length; i++) {

		current_addr += 4;

		if (sbp2util_read_quadlet(hi, LOCAL_BUS | node_id, current_addr, &current_quadlet)) {
			SBP2_DEBUG("sbp2: Error reading at address 0x%08x%08x - bad status", 
				   (unsigned int) ((current_addr) >> 32), (unsigned int) ((current_addr) & 0xffffffff));
			return(-EIO);   
		}

		/* 
		 * Check for unit_spec_ID tag 
		 */
		if ((be32_to_cpu(current_quadlet) >> 24) == SBP2_UNIT_SPEC_ID_KEY) {
			unit_spec_id_data = current_quadlet;
			SBP2_DEBUG("sbp2: Node %x, unit spec id = %x", (LOCAL_BUS | node_id), 
				   (unsigned int) be32_to_cpu(unit_spec_id_data));
		}

		/* 
		 * Check for unit_sw_version tag 
		 */
		if ((be32_to_cpu(current_quadlet) >> 24) == SBP2_UNIT_SW_VERSION_KEY) {
			unit_sw_ver_data = current_quadlet;
			SBP2_DEBUG("sbp2: Node %x, unit sw version = %x", (LOCAL_BUS | node_id), 
				   (unsigned int) be32_to_cpu(unit_sw_ver_data));
		}
	}

	/*
	 * Validate unit spec id and unit sw ver to see if this is an SBP-2 device
	 */
	if ((be32_to_cpu(unit_spec_id_data) != SBP2_UNIT_SPEC_ID_ENTRY) ||
	    (be32_to_cpu(unit_sw_ver_data) != SBP2_SW_VERSION_ENTRY)) {

		/*
		 * Not an sbp2 device
		 */
		return(-ENXIO);
	}

	/*
	 * This device is a valid SBP-2 device
	 */
	SBP2_ERR("sbp2: Node %x, Found SBP-2 device", (LOCAL_BUS | node_id));
	return(0);
}

/*
 * This function removes (cleans-up after) any unvalidated sbp2 devices
 */
static void sbp2_remove_unvalidated_devices(struct sbp2scsi_host_info *hi)
{
	int i;

	/*
	 * Loop through and free any unvalidated scsi id instance data structures
	 */
	for (i=0; i<SBP2SCSI_MAX_SCSI_IDS; i++) {
		if (hi->scsi_id[i]) {
			if (!hi->scsi_id[i]->validated) {

				/*
				 * Complete any pending commands (which were not already completed)
				 */
				if (hi->scsi_id[i]->Current_SCpnt) {
					sbp2scsi_complete_command(hi, hi->scsi_id[i], SBP2_SCSI_STATUS_SELECTION_TIMEOUT);
				}

				/* 
				 * Clean up any other structures
				 */
				kfree(hi->scsi_id[i]);
				hi->scsi_id[i] = NULL;
				SBP2_DEBUG("sbp2: Unvalidated SBP-2 device removed, SCSI ID = %x", (unsigned int) i);
			}
		}
	}

	return;
}

/*
 * This function is our reset handler. It is run out of a thread, since we get 
 * notified of a bus reset from a bh (or interrupt).
 */
static void sbp2_bus_reset_handler(void *context)
{
	struct sbp2scsi_host_info *hi = context;
	quadlet_t signature_data;
	int i;
	unsigned long flags;

	SBP2_DEBUG("sbp2: sbp2_bus_reset_handler");

	/*
	 * TODO. Check and keep track of generation number of all requests, in case a
	 * bus reset occurs while trying to find and login to SBP-2 devices.
	 */

	/*
	 * First thing to do. Invalidate all SBP-2 devices. This is needed so that
	 * we stop sending down I/O requests to the device, and also so that we can
	 * figure out which devices have disappeared after a bus reset.
	 */
	for (i=0; i<SBP2SCSI_MAX_SCSI_IDS; i++) {
		if (hi->scsi_id[i]) {
			hi->scsi_id[i]->validated = 0;
		}
	}

	/*
	 * Give the sbp2 devices a little time to recover after the bus reset
	 */
	current->state = TASK_INTERRUPTIBLE;
	schedule_timeout(HZ/2);		/* 1/2 second delay */

	/*
	 * Spit out what we know from the host
	 */
	SBP2_DEBUG("host: node_count = %x", (unsigned int) hi->host->node_count);
	SBP2_DEBUG("host: selfid_count = %x", (unsigned int) hi->host->selfid_count);
	SBP2_DEBUG("host: node_id = %x", (unsigned int) hi->host->node_id);
	SBP2_DEBUG("host: irm_id = %x", (unsigned int) hi->host->irm_id);
	SBP2_DEBUG("host: busmgr_id = %x", (unsigned int) hi->host->busmgr_id);
	SBP2_DEBUG("host: is_root = %x", (unsigned int) hi->host->is_root);
	SBP2_DEBUG("host: is_cycmst = %x", (unsigned int) hi->host->is_cycmst);
	SBP2_DEBUG("host: is_irm = %x", (unsigned int) hi->host->is_irm);
	SBP2_DEBUG("host: is_busmgr = %x", (unsigned int) hi->host->is_busmgr);

	/*
	 * Let's try and figure out which devices out there are SBP-2 devices! Loop through all 
	 * nodes out there.
	 */
	for (i=0; i<hi->host->node_count; i++) {

		/*
		 * Don't read from ourselves!
		 */
		if (i != ((hi->host->node_id) & NODE_MASK)) {

			/*
			 * Try and send a request for a config rom signature. This is expected to fail for
			 * some nodes, as they might be repeater phys or not be initialized.
			 */
			if (!sbp2util_read_quadlet(hi, LOCAL_BUS | i, CONFIG_ROM_SIGNATURE_ADDRESS, &signature_data)) {

				if (be32_to_cpu(signature_data) == IEEE1394_CONFIG_ROM_SIGNATURE) {

					/*
					 * Hey, we've got a valid responding IEEE1394 node. Need to now see if it's an SBP-2 device
					 */
					if (!sbp2_check_device(hi, i)) {

						/*
						 * Found an SBP-2 device. Now, actually start the device.
						 */
						sbp2_start_device(hi, i);
					}
				}
			}
		}
	}

	/*
	 * This code needs protection
	 */
	spin_lock_irqsave(&hi->sbp2_command_lock, flags);

	/*
	 * Ok, we've discovered and re-validated all SBP-2 devices out there. Let's remove structures of all
	 * devices not re-validated (meaning they've been removed).
	 */
	sbp2_remove_unvalidated_devices(hi);

	/*
	 * Now, note that the bus reset is complete (finally!)
	 */
	hi->bus_reset_in_progress = 0;

	/*
	 * Let's run through any pending commands and re-execute them 
	 */
	for (i=0; i<SBP2SCSI_MAX_SCSI_IDS; i++) {
		if (hi->scsi_id[i]) {
			if (hi->scsi_id[i]->Current_SCpnt) {
				if (sbp2_send_command(hi, hi->scsi_id[i])) {
					SBP2_ERR("sbp2: Error sending SCSI command");
					sbp2scsi_complete_command(hi, hi->scsi_id[i], SBP2_SCSI_STATUS_SELECTION_TIMEOUT);
				}
			}
		}
	}
	spin_unlock_irqrestore(&hi->sbp2_command_lock, flags);

	return;
}


/*
 * This is called from the host's bh when a bus reset is complete. We wake up our detection thread
 * to deal with the reset
 */
static void sbp2_host_reset(struct hpsb_host *host)
{
	unsigned long flags;
	struct sbp2scsi_host_info *hi;
	int i;

	SBP2_ERR("sbp2: IEEE-1394 bus reset");
	spin_lock_irqsave(&sbp2_host_info_lock, flags);
	hi = sbp2_find_host_info(host);

	if (hi != NULL) {

		/*
		 * Wake up our detection thread, only if it's not already handling a reset
		 */
		if (!hi->bus_reset_in_progress) {

			/*
			 * First thing to do. Invalidate all SBP-2 devices. This is needed so that
			 * we stop sending down I/O requests to the device, and also so that we can
			 * figure out which devices have disappeared after a bus reset.
			 */
			for (i=0; i<SBP2SCSI_MAX_SCSI_IDS; i++) {
				if (hi->scsi_id[i]) {
					hi->scsi_id[i]->validated = 0;
				}
			}

			hi->bus_reset_in_progress = 1;

			wake_up(&hi->sbp2_detection_wait);
		}
	}
	spin_unlock_irqrestore(&sbp2_host_info_lock, flags);
	return;
}

/*
 * This function deals with physical dma write requests (for adapters that do not support
 * physical dma in hardware).
 */
#ifdef SBP2_EMULATE_PHYSICAL_DMA_SUPPORT
static int sbp2_handle_physdma_write(struct hpsb_host *host, int nodeid, quadlet_t *data,
				     u64 addr, unsigned int length)
{

	/*
	 * Manually put the data in the right place.
	 */
	memcpy(bus_to_virt((u32)addr), data, length);
	return(RCODE_COMPLETE);
}
#endif

/*
 * This function deals with physical dma read requests (for adapters that do not support
 * physical dma in hardware).
 */
#ifdef SBP2_EMULATE_PHYSICAL_DMA_SUPPORT
static int sbp2_handle_physdma_read(struct hpsb_host *host, int nodeid, quadlet_t *data,
				    u64 addr, unsigned int length)
{

	/*
	 * Grab data from memory and send a read response.
	 */
	memcpy(data, bus_to_virt((u32)addr), length);
	return(RCODE_COMPLETE);
}
#endif


/**************************************
 * SBP-2 protocol related section
 **************************************/

/*
 * This function is called in order to login to a particular SBP-2 device, after a bus reset
 */
static int sbp2_login_device(struct sbp2scsi_host_info *hi, struct scsi_id_instance_data *scsi_id) 
{
	quadlet_t data[2];

	SBP2_DEBUG("sbp2: sbp2_login_device");

	/*
	 * Make sure that we've gotten ahold of the sbp2 management agent address. Also figure out the
	 * command set being used (SCSI or RBC).
	 */
	if (sbp2_parse_unit_directory(hi, scsi_id)) {
		SBP2_ERR("sbp2: Error while parsing sbp2 unit directory");
		return(-EIO);
	}

	/*
	 * Set-up login ORB
	 */
	scsi_id->login_orb.password = 0x0ULL;
	scsi_id->login_orb.login_response_lo = virt_to_bus(&scsi_id->login_response);
	scsi_id->login_orb.login_response_hi = (hi->host->node_id) << 16;
	scsi_id->login_orb.lun = 0x0;	  /* temporary, until we support multiple luns */
	scsi_id->login_orb.function = LOGIN_REQUEST;
	scsi_id->login_orb.reconnect = 0x0;
	scsi_id->login_orb.reserved = 0x0;
	scsi_id->login_orb.x = 0x1;
	scsi_id->login_orb.rq_fmt = 0x0;
	scsi_id->login_orb.n = 0x1;
	scsi_id->login_orb.login_response_length = sizeof(struct sbp2_login_response);
	scsi_id->login_orb.password_length = 0x0;
	scsi_id->login_orb.status_FIFO_lo = (u32)SBP2_STATUS_FIFO_ADDRESS;
	scsi_id->login_orb.status_FIFO_hi = ((hi->host->node_id) << 16) | ((u32)(SBP2_STATUS_FIFO_ADDRESS >> 32));

	/*
	 * Byte swap ORB if necessary
	 */
	sbp2util_cpu_to_be32_buffer(&scsi_id->login_orb, sizeof(struct sbp2_login_orb));

	/*
	 * Initialize login response and status fifo
	 */
	memset(&scsi_id->login_response, 0, sizeof(struct sbp2_login_response));
	memset(&scsi_id->status_block, 0, sizeof(struct sbp2_status_block));

	/*
	 * Ok, let's write to the target's management agent register
	 */
	data[0] = (hi->host->node_id) << 16;
	data[1] = virt_to_bus(&scsi_id->login_orb);
	sbp2util_cpu_to_be32_buffer(data, 8);

	hpsb_write(hi->host, LOCAL_BUS | scsi_id->node_id, scsi_id->sbp2_management_agent_addr, data, 8);

	/*
	 * Wait for login status...
	 */
	interruptible_sleep_on_timeout(&scsi_id->sbp2_login_wait, HZ);		/* 1 second timeout */

	/*
	 * Check status
	 */
	if (scsi_id->status_block.resp ||
	    scsi_id->status_block.d ||
	    scsi_id->status_block.sbp_status) {

		SBP2_ERR("sbp2: Error logging into SBP-2 device - login failed");
		return(-EIO);
	}

	/*
	 * Grab our command block agent address from the login response
	 */
	SBP2_DEBUG("sbp2: command_block_agent_hi = %x", (unsigned int)be32_to_cpu(scsi_id->login_response.command_block_agent_hi));
	SBP2_DEBUG("sbp2: command_block_agent_lo = %x", (unsigned int)be32_to_cpu(scsi_id->login_response.command_block_agent_lo));

	scsi_id->sbp2_command_block_agent_addr = ((u64)(be32_to_cpu(scsi_id->login_response.command_block_agent_hi)) << 32);
	scsi_id->sbp2_command_block_agent_addr |= ((u64)(be32_to_cpu(scsi_id->login_response.command_block_agent_lo)));
	scsi_id->sbp2_command_block_agent_addr &= 0x0000ffffffffffffULL;

	SBP2_DEBUG("sbp2: sbp2_command_block_agent_addr = %x", (unsigned int)scsi_id->sbp2_command_block_agent_addr);

	SBP2_ERR("sbp2: Logged into SBP-2 device");

	return(0);

}

/*
 * This function is called in order to logout from a particular SBP-2 device, usually called during driver
 * unload
 */
static int sbp2_logout_device(struct sbp2scsi_host_info *hi, struct scsi_id_instance_data *scsi_id) 
{
	quadlet_t data[2];

	SBP2_DEBUG("sbp2: sbp2_logout_device");

	/*
	 * Set-up logout ORB
	 */
	scsi_id->logout_orb.reserved1 = 0x0;
	scsi_id->logout_orb.reserved2 = 0x0;
	scsi_id->logout_orb.login_ID = be32_to_cpu(scsi_id->login_response.login_ID);
	scsi_id->logout_orb.function = LOGOUT_REQUEST;
	scsi_id->logout_orb.reserved3 = 0x0;
	scsi_id->logout_orb.rq_fmt = 0x0;
	scsi_id->logout_orb.n = 0x1;
	scsi_id->logout_orb.reserved4 = 0;
	scsi_id->logout_orb.status_FIFO_lo = (u32)SBP2_STATUS_FIFO_ADDRESS;
	scsi_id->logout_orb.status_FIFO_hi = ((hi->host->node_id) << 16) | ((u32)(SBP2_STATUS_FIFO_ADDRESS >> 32));

	/*
	 * Byte swap ORB if necessary
	 */
	sbp2util_cpu_to_be32_buffer(&scsi_id->logout_orb, sizeof(struct sbp2_logout_orb));

	/*
	 * Ok, let's write to the target's management agent register
	 */
	data[0] = (hi->host->node_id) << 16;
	data[1] = virt_to_bus(&scsi_id->logout_orb);
	sbp2util_cpu_to_be32_buffer(data, 8);

	hpsb_write(hi->host, LOCAL_BUS | scsi_id->node_id, scsi_id->sbp2_management_agent_addr, data, 8);

	/*
	 * Wait for device to logout... 
	 */
	interruptible_sleep_on_timeout(&scsi_id->sbp2_login_wait, HZ);		/* 1 second timeout */

	SBP2_ERR("sbp2: Logged out of SBP-2 device");

	return(0);

}

/*
 * This function is called in order to reconnect to a particular SBP-2 device, after a bus reset
 */
static int sbp2_reconnect_device(struct sbp2scsi_host_info *hi, struct scsi_id_instance_data *scsi_id) 
{
	quadlet_t data[2];

	SBP2_DEBUG("sbp2: sbp2_reconnect_device");

	/*
	 * Set-up reconnect ORB
	 */
	scsi_id->reconnect_orb.reserved1 = 0x0;
	scsi_id->reconnect_orb.reserved2 = 0x0;
	scsi_id->reconnect_orb.login_ID = be32_to_cpu(scsi_id->login_response.login_ID);
	scsi_id->reconnect_orb.function = RECONNECT_REQUEST;
	scsi_id->reconnect_orb.reserved3 = 0x0;
	scsi_id->reconnect_orb.rq_fmt = 0x0;
	scsi_id->reconnect_orb.n = 0x1;
	scsi_id->reconnect_orb.reserved4 = 0x0;
	scsi_id->reconnect_orb.status_FIFO_lo = (u32)SBP2_STATUS_FIFO_ADDRESS;
	scsi_id->reconnect_orb.status_FIFO_hi = ((hi->host->node_id) << 16) | ((u32)(SBP2_STATUS_FIFO_ADDRESS >> 32));

	/*
	 * Byte swap ORB if necessary
	 */
	sbp2util_cpu_to_be32_buffer(&scsi_id->reconnect_orb, sizeof(struct sbp2_reconnect_orb));

	/*
	 * Initialize status fifo
	 */
	memset(&scsi_id->status_block, 0, sizeof(struct sbp2_status_block));

	/*
	 * Ok, let's write to the target's management agent register
	 */
	data[0] = (hi->host->node_id) << 16;
	data[1] = virt_to_bus(&scsi_id->reconnect_orb);
	sbp2util_cpu_to_be32_buffer(data, 8);

	hpsb_write(hi->host, LOCAL_BUS | scsi_id->node_id, scsi_id->sbp2_management_agent_addr, data, 8);

	/*
	 * Wait for reconnect status...
	 */
	interruptible_sleep_on_timeout(&scsi_id->sbp2_login_wait, HZ);		/* 1 second timeout */

	/*
	 * Check status
	 */
	if (scsi_id->status_block.resp ||
	    scsi_id->status_block.d ||
	    scsi_id->status_block.sbp_status) {

		SBP2_ERR("sbp2: Error reconnecting to SBP-2 device - reconnect failed");
		return(-EIO);
	}

	SBP2_ERR("sbp2: Reconnected to SBP-2 device");

	return(0);

}

/*
 * This function is called in order to set the busy timeout (number of retries to attempt) on the sbp2 device. 
 */
static int sbp2_set_busy_timeout(struct sbp2scsi_host_info *hi, struct scsi_id_instance_data *scsi_id)
{      
	quadlet_t data;

	SBP2_DEBUG("sbp2: sbp2_set_busy_timeout");

	/*
	 * Ok, let's write to the target's busy timeout register
	 */
	data = cpu_to_be32(SBP2_BUSY_TIMEOUT_VALUE);

	if (hpsb_write(hi->host, LOCAL_BUS | scsi_id->node_id, SBP2_BUSY_TIMEOUT_ADDRESS, &data, 4)) {
		SBP2_ERR("sbp2: sbp2_set_busy_timeout error");
	}

	return(0);
}

/*
 * This function is called to parse sbp2 device's config rom unit directory. Used to determine
 * things like sbp2 management agent offset, and command set used (SCSI or RBC). 
 */
static int sbp2_parse_unit_directory(struct sbp2scsi_host_info *hi, struct scsi_id_instance_data *scsi_id)
{
	quadlet_t unit_directory_length, unit_directory_data;
	u64 unit_directory_addr;
	u32 i;

	SBP2_DEBUG("sbp2: sbp2_parse_unit_directory");

	if (sbp2util_unit_directory(hi, LOCAL_BUS | scsi_id->node_id, &unit_directory_addr)) {
		SBP2_DEBUG("sbp2: Error reading unit directory address - bad status");
		return(-EIO);   
	}

	/*
	 * Read the size of the unit directory
	 */
	if (sbp2util_read_quadlet(hi, LOCAL_BUS | scsi_id->node_id, unit_directory_addr, 
				  &unit_directory_length)) {
		SBP2_DEBUG("sbp2: Error reading unit directory length - bad status");
		return(-EIO);   
	}

	unit_directory_length = ((be32_to_cpu(unit_directory_length)) >> 16);

	/*
	 * Now, sweep through the unit directory looking for the management agent offset
	 * Give up if we hit any error or somehow miss it...
	 */
	for (i=0; i<unit_directory_length; i++) {

		if (sbp2util_read_quadlet(hi, LOCAL_BUS | scsi_id->node_id, unit_directory_addr + (i<<2) + 4, 
					  &unit_directory_data)) {
			SBP2_DEBUG("sbp2: Error reading unit directory - bad status");
			return(-EIO);
		}

		/* 
		 * Handle different fields in the unit directory, based on keys
		 */
		unit_directory_data = be32_to_cpu(unit_directory_data);
		switch (unit_directory_data >> 24) {
			
			case SBP2_CSR_OFFSET_KEY:

				/*
				 * Save off the management agent address
				 */
				scsi_id->sbp2_management_agent_addr = CONFIG_ROM_INITIAL_MEMORY_SPACE + 
								      ((unit_directory_data & 0x00ffffff) << 2);

				SBP2_DEBUG("sbp2: sbp2_management_agent_addr = %x", (unsigned int) scsi_id->sbp2_management_agent_addr);
				break;

			case SBP2_COMMAND_SET_SPEC_ID_KEY:

				/*
				 * Command spec organization
				 */
				scsi_id->sbp2_command_set_spec_id = unit_directory_data & 0xffffff;
				SBP2_DEBUG("sbp2: sbp2_command_set_spec_id = %x", (unsigned int) scsi_id->sbp2_command_set_spec_id);
				break;

			case SBP2_COMMAND_SET_KEY:

				/*
				 * Command set used by sbp2 device
				 */
				scsi_id->sbp2_command_set = unit_directory_data & 0xffffff;
				SBP2_DEBUG("sbp2: sbp2_command_set = %x", (unsigned int) scsi_id->sbp2_command_set);
				break;

			case SBP2_UNIT_CHARACTERISTICS_KEY:

				/*
				 * Unit characterisitcs (orb related stuff that I'm not yet paying attention to)
				 */
				scsi_id->sbp2_unit_characteristics = unit_directory_data & 0xffffff;
				SBP2_DEBUG("sbp2: sbp2_unit_characteristics = %x", (unsigned int) scsi_id->sbp2_unit_characteristics);
				break;

			case SBP2_DEVICE_TYPE_AND_LUN_KEY:

				/*
				 * Device type and lun (used for detemining type of sbp2 device)
				 */
				scsi_id->sbp2_device_type_and_lun = unit_directory_data & 0xffffff;
				SBP2_DEBUG("sbp2: sbp2_device_type_and_lun = %x", (unsigned int) scsi_id->sbp2_device_type_and_lun);
				break;

			case SBP2_UNIT_SPEC_ID_KEY:

				/*
				 * Unit spec id (used for protocol detection)
				 */
				scsi_id->sbp2_unit_spec_id = unit_directory_data & 0xffffff;
				SBP2_DEBUG("sbp2: sbp2_unit_spec_id = %x", (unsigned int) scsi_id->sbp2_unit_spec_id);
				break;

			case SBP2_UNIT_SW_VERSION_KEY:

				/*
				 * Unit sw version (used for protocol detection) 
				 */
				scsi_id->sbp2_unit_sw_version = unit_directory_data & 0xffffff;
				SBP2_DEBUG("sbp2: sbp2_unit_sw_version = %x", (unsigned int) scsi_id->sbp2_unit_sw_version);
				break;

			default:
				break;
		}

	}

	return(0);
}

/*
 * This function is called in order to enable unsolicited status on the sbp2 device. 
 */
#ifdef SBP2_ENABLE_UNSOLICITED_STATUS
static int sbp2_enable_unsolicited_status(struct sbp2scsi_host_info *hi, struct scsi_id_instance_data *scsi_id)
{
	quadlet_t data;

	SBP2_DEBUG("sbp2: sbp2_enable_unsolicited_status");

	/*
	 * Let's write to the target's unsolicited status enable address
	 */
	data = cpu_to_be32(SBP2_UNSOLICITED_STATUS_VALUE);

	if (hpsb_write(hi->host, LOCAL_BUS | scsi_id->node_id, 
		       scsi_id->sbp2_command_block_agent_addr + SBP2_UNSOLICITED_STATUS_ENABLE_OFFSET, 
		       &data, 4)) {
		SBP2_ERR("sbp2: sbp2_enable_unsolicited_status error");
	}

	return(0);
}
#endif

/*
 * This function is called in order to determine the max speed and packet size we can use in our ORBs. 
 */
static int sbp2_max_speed_and_size(struct sbp2scsi_host_info *hi, struct scsi_id_instance_data *scsi_id)
{
	quadlet_t node_options, max_rec;
	u8 speed_code;

	SBP2_DEBUG("sbp2: sbp2_max_speed_and_size");

	/*
	 * Get speed code from internal host structure. There should be a better way to obtain this.
	 */
	speed_code = hi->host->speed_map[(hi->host->node_id & NODE_MASK) * 64 + (scsi_id->node_id & NODE_MASK)];

	switch (speed_code) {
		case SPEED_S100:
			scsi_id->speed_code = SPEED_S100;
			scsi_id->max_payload_size = MAX_PAYLOAD_S100;
			SBP2_ERR("sbp2: SBP-2 device max speed S100 and payload 512 bytes"); 
			break;
		case SPEED_S200:
			scsi_id->speed_code = SPEED_S200;
			scsi_id->max_payload_size = MAX_PAYLOAD_S200;
			SBP2_ERR("sbp2: SBP-2 device max speed S200 and payload 1KB"); 
			break;
		case SPEED_S400:
			scsi_id->speed_code = SPEED_S400;
			scsi_id->max_payload_size = MAX_PAYLOAD_S400;
			SBP2_ERR("sbp2: SBP-2 device max speed S400 and payload 2KB"); 
			break;
		default:
			scsi_id->speed_code = SPEED_S100;
			scsi_id->max_payload_size = MAX_PAYLOAD_S100;
			SBP2_ERR("sbp2: Undefined speed: Using SBP-2 device max speed S100 and payload 512 bytes"); 
			break;
	}

	/*
	 * Device speed/payload hacks go here...
	 */
	if (scsi_id->speed_code > SBP2_MAX_SPEED_ALLOWED) {
		scsi_id->speed_code = SBP2_MAX_SPEED_ALLOWED;
		SBP2_ERR("sbp2: Reducing SBP-2 max speed allowed (%x)", SBP2_MAX_SPEED_ALLOWED); 
	}

	if (scsi_id->max_payload_size > SBP2_MAX_PAYLOAD_ALLOWED) {
		scsi_id->max_payload_size = SBP2_MAX_PAYLOAD_ALLOWED;
		SBP2_ERR("sbp2: Reducing SBP-2 max payload allowed (%x)", SBP2_MAX_PAYLOAD_ALLOWED); 
	}

	/*
	 * Finally, check the adapter's capabilities to further bump down our max payload size
	 * if necessary. For instance, TILynx may not support the default max payload at a 
	 * particular speed.
	 */
	if (!hpsb_read(hi->host, hi->host->node_id | LOCAL_BUS, CONFIG_ROM_NODE_OPTIONS, &node_options, 4)) {

		/* 
		 * Grab max_rec (max payload = 2 ^ (max_rec+1)) from node options. Sbp2 max payload is 
		 * defined as 2 ^ (max_pay+2)... so, have to subtract one from max rec for comparison...
		 * confusing, eh?   ;-)
		 */
		max_rec = (be32_to_cpu(node_options) & 0x0000f000) >> 12;
		if (scsi_id->max_payload_size > (max_rec - 1)) {
			scsi_id->max_payload_size = (max_rec - 1);
			SBP2_ERR("sbp2: Reducing SBP-2 max payload allowed (%x)", (max_rec - 1)); 
		}

	}

	return(0);
}

/*
 * This function is called in order to perform a SBP-2 agent reset. 
 */
static int sbp2_agent_reset(struct sbp2scsi_host_info *hi, struct scsi_id_instance_data *scsi_id, u32 flags) 
{
	struct sbp2_request_packet *agent_reset_request_packet;

	SBP2_DEBUG("sbp2: sbp2_agent_reset");

	/*
	 * Ok, let's write to the target's management agent register
	 */
	agent_reset_request_packet = sbp2util_allocate_write_request_packet(hi, LOCAL_BUS | scsi_id->node_id,
									    scsi_id->sbp2_command_block_agent_addr + SBP2_AGENT_RESET_OFFSET,
									    0, ntohl(SBP2_AGENT_RESET_DATA));

	if (!agent_reset_request_packet) {
		SBP2_ERR("sbp2: sbp2util_allocate_write_request_packet failed");
		return(-EIO);
	}

	if (!hpsb_send_packet(agent_reset_request_packet->packet)) {
		SBP2_ERR("sbp2: hpsb_send_packet failed");
		sbp2util_free_request_packet(agent_reset_request_packet); 
		return(-EIO);
	}

	if (!(flags & SBP2_SEND_NO_WAIT)) {
		down(&agent_reset_request_packet->packet->state_change);
		down(&agent_reset_request_packet->packet->state_change);
	}

	return(0);

}

/*
 * This function is called in order to begin a regular SBP-2 command. 
 */
static int sbp2_send_command(struct sbp2scsi_host_info *hi, struct scsi_id_instance_data *scsi_id)
{
	Scsi_Cmnd *SCpnt = scsi_id->Current_SCpnt;
	unchar *cmd = (unchar *) SCpnt->cmnd;
	struct scatterlist *sgpnt;
	int i;
	struct sbp2_request_packet *command_request_packet;
	u32 device_type;

	SBP2_DEBUG("sbp2: sbp2_send_command");
	SBP2_DEBUG("sbp2: SCSI command = %02x", *cmd);
	SBP2_DEBUG("sbp2: SCSI command length = %x", COMMAND_SIZE(*cmd));
	SBP2_DEBUG("sbp2: SCSI data transfer direction = %x", sbp2scsi_direction_table[*cmd]);

	/*
	 * Set-up our command ORB..
	 *
	 * NOTE: We're doing unrestricted page tables (s/g), as this is best performance 
	 * (at least with the devices I have). This means that data_size becomes the number 
	 * of s/g elements, and page_size should be zero (for unrestricted).
	 */
	scsi_id->command_orb.next_ORB_hi = 0xffffffff;
	scsi_id->command_orb.next_ORB_lo = 0xffffffff;
	scsi_id->command_orb.page_size = 0x0;			/* use unrestricted page tables (s/g) */
	scsi_id->command_orb.max_payload = scsi_id->max_payload_size;   
	scsi_id->command_orb.spd = scsi_id->speed_code; 
	scsi_id->command_orb.r = 0x0;
	scsi_id->command_orb.rq_fmt = 0x0;
	scsi_id->command_orb.n = 0x1;

	/*
	 * Set-up our pagetable stuff...
	 */
	if (sbp2scsi_direction_table[*cmd] == ORB_DIRECTION_NO_DATA_TRANSFER) {

		SBP2_DEBUG("sbp2: No data transfer");

		/*
		 * Handle no data transfer
		 */
		scsi_id->command_orb.p = 0x0;			/* use page tables (s/g) */
		scsi_id->command_orb.data_size = 0x0;		/* number of page table (s/g) elements */
		scsi_id->command_orb.data_descriptor_hi = 0xffffffff;
		scsi_id->command_orb.data_descriptor_lo = 0xffffffff;
		scsi_id->command_orb.d = sbp2scsi_direction_table[*cmd];

	} else if (SCpnt->use_sg) {

		SBP2_DEBUG("sbp2: Use scatter/gather");

		/*
		 * Get s/g list from request
		 */
		sgpnt = (struct scatterlist *) SCpnt->request_buffer;

		/*
		 * Special case if only one element
		 */
		if (SCpnt->use_sg == 1) {

			SBP2_DEBUG("sbp2: Only one s/g element");

			scsi_id->command_orb.p = 0x0;
			scsi_id->command_orb.data_descriptor_hi = (hi->host->node_id) << 16;
			scsi_id->command_orb.data_descriptor_lo = virt_to_bus(sgpnt[0].address);
			scsi_id->command_orb.data_size = sgpnt[0].length;
			scsi_id->command_orb.d = sbp2scsi_direction_table[*cmd];

		} else {

			scsi_id->command_orb.p = 0x1;				/* use page tables (s/g) */
			scsi_id->command_orb.data_size = SCpnt->use_sg; /* number of page table (s/g) elements */
			scsi_id->command_orb.data_descriptor_hi = (hi->host->node_id) << 16;
			scsi_id->command_orb.data_descriptor_lo = virt_to_bus(scsi_id->scatter_gather_element);
			scsi_id->command_orb.d = sbp2scsi_direction_table[*cmd];

			/*
			 * Loop through and fill out our sbp-2 page tables
			 */
			for (i=0; i<SCpnt->use_sg; i++) {
				scsi_id->scatter_gather_element[i].segment_length = sgpnt[i].length;
				scsi_id->scatter_gather_element[i].segment_base_hi = 0x0;
				scsi_id->scatter_gather_element[i].segment_base_lo = virt_to_bus(sgpnt[i].address);
			}

			/*
			 * Byte swap page tables if necessary
			 */
			sbp2util_cpu_to_be32_buffer(scsi_id->scatter_gather_element, 
						    (sizeof(struct sbp2_unrestricted_page_table)) * scsi_id->command_orb.data_size);

		}

	} else {

		SBP2_DEBUG("sbp2: No scatter/gather");

		/*
		 * Handle case where we get a command w/o s/g enabled
		 */
		scsi_id->command_orb.p = 0x0;
		scsi_id->command_orb.data_descriptor_hi = (hi->host->node_id) << 16;
		scsi_id->command_orb.data_descriptor_lo = virt_to_bus(SCpnt->request_buffer);
		scsi_id->command_orb.data_size = SCpnt->request_bufflen;
		scsi_id->command_orb.d = sbp2scsi_direction_table[*cmd];

		/*
		 * Sanity, in case our direction table is not up-to-date
		 */
		if (!SCpnt->request_bufflen) {
			scsi_id->command_orb.data_descriptor_hi = 0xffffffff;
			scsi_id->command_orb.data_descriptor_lo = 0xffffffff;
			scsi_id->command_orb.d = ORB_DIRECTION_NO_DATA_TRANSFER;
		}

	}

	/*
	 * Byte swap command ORB if necessary
	 */
	sbp2util_cpu_to_be32_buffer(&scsi_id->command_orb, sizeof(struct sbp2_command_orb));

	/*
	 * Put our scsi command in the command ORB
	 */
	memset(scsi_id->command_orb.cdb, 0, 12);
	memcpy(scsi_id->command_orb.cdb, cmd, COMMAND_SIZE(*cmd));

	/*
	 * Update our cdb if necessary (to handle sbp2 RBC command set differences).
	 * This is where the command set hacks go!   =)
	 */
	device_type = (scsi_id->sbp2_device_type_and_lun & 0x00ff0000) >> 16;
	if ((device_type == TYPE_DISK) ||
	    (device_type == TYPE_SDAD)) {
		sbp2_check_sbp2_command(scsi_id->command_orb.cdb);
	}

	/*
	 * Initialize status fifo
	 */
	memset(&scsi_id->status_block, 0, sizeof(struct sbp2_status_block));

	/*
	 * Ok, let's write to the target's management agent register
	 */
	command_request_packet = sbp2util_allocate_write_request_packet(hi, LOCAL_BUS | scsi_id->node_id,
									scsi_id->sbp2_command_block_agent_addr + SBP2_ORB_POINTER_OFFSET,
									8, 0);

	if (!command_request_packet) {
		SBP2_ERR("sbp2: sbp2util_allocate_write_request_packet failed");
		return(-EIO);
	}

	command_request_packet->packet->data[0] = (hi->host->node_id) << 16;
	command_request_packet->packet->data[1] = virt_to_bus(&scsi_id->command_orb);
	sbp2util_cpu_to_be32_buffer(command_request_packet->packet->data, 8);

	if (!hpsb_send_packet(command_request_packet->packet)) {
		SBP2_ERR("sbp2: hpsb_send_packet failed");
		sbp2util_free_request_packet(command_request_packet); 
		return(-EIO);
	}

	return(0);
}

/*
 * This function deals with command set differences between Linux scsi command set and sbp2 RBC
 * command set.
 */
static void sbp2_check_sbp2_command(unchar *cmd)
{
	unchar new_cmd[16];

	SBP2_DEBUG("sbp2: sbp2_check_sbp2_command");

	switch (*cmd) {
		
		case READ_6:

			SBP2_DEBUG("sbp2: Convert READ_6 to READ_10");

			/*
			 * Need to turn read_6 into read_10
			 */
			new_cmd[0] = 0x28;
			new_cmd[1] = (cmd[1] & 0xe0);
			new_cmd[2] = 0x0;
			new_cmd[3] = (cmd[1] & 0x1f);
			new_cmd[4] = cmd[2];
			new_cmd[5] = cmd[3];
			new_cmd[6] = 0x0;
			new_cmd[7] = 0x0;
			new_cmd[8] = cmd[4];
			new_cmd[9] = cmd[5];

			memcpy(cmd, new_cmd, 10);

			break;

		case WRITE_6:

			SBP2_DEBUG("sbp2: Convert WRITE_6 to WRITE_10");

			/*
			 * Need to turn write_6 into write_10
			 */
			new_cmd[0] = 0x2a;
			new_cmd[1] = (cmd[1] & 0xe0);
			new_cmd[2] = 0x0;
			new_cmd[3] = (cmd[1] & 0x1f);
			new_cmd[4] = cmd[2];
			new_cmd[5] = cmd[3];
			new_cmd[6] = 0x0;
			new_cmd[7] = 0x0;
			new_cmd[8] = cmd[4];
			new_cmd[9] = cmd[5];

			memcpy(cmd, new_cmd, 10);

			break;

		case MODE_SENSE:
		case MODE_SELECT:

			/*
			 * TODO. Probably need to change mode sense/select to 10 byte versions
			 */

		default:
			break;
	}

	return;
}

/*
 * Translates SBP-2 status into SCSI sense data for check conditions
 */
static void sbp2_status_to_sense_data(unchar *sbp2_status, unchar *sense_data)
{
	SBP2_DEBUG("sbp2: sbp2_status_to_sense_data");

	/*
	 * Ok, it's pretty ugly...   ;-)
	 */
	sense_data[0] = 0x70;
	sense_data[1] = 0x0;
	sense_data[2] = sbp2_status[9];
	sense_data[3] = sbp2_status[12];
	sense_data[4] = sbp2_status[13];
	sense_data[5] = sbp2_status[14];
	sense_data[6] = sbp2_status[15];
	sense_data[7] = 10;
	sense_data[8] = sbp2_status[16];
	sense_data[9] = sbp2_status[17];
	sense_data[10] = sbp2_status[18];
	sense_data[11] = sbp2_status[19];
	sense_data[12] = sbp2_status[10];
	sense_data[13] = sbp2_status[11];
	sense_data[14] = sbp2_status[20];
	sense_data[15] = sbp2_status[21];

	return;
}

/*
 * This function is called after a command is completed, in order to do any necessary SBP-2
 * response data translations for the SCSI stack
 */
static void sbp2_check_sbp2_response(Scsi_Cmnd *SCpnt)
{
	u8 *scsi_buf = SCpnt->request_buffer;

	SBP2_DEBUG("sbp2: sbp2_check_sbp2_response");

	switch (SCpnt->cmnd[0]) {
		
		case INQUIRY:

			SBP2_DEBUG("sbp2: Check Inquiry data");

			/*
			 * Check for Simple Direct Access Device and change it to TYPE_DISK
			 */
			if ((scsi_buf[0] & 0x1f) == TYPE_SDAD) {
				SBP2_DEBUG("sbp2: Changing TYPE_SDAD to TYPE_DISK");
				scsi_buf[0] &= 0xe0;
			}

			/*
			 * Fix ansi revision and response data format
			 */
			scsi_buf[2] |= 2;
			scsi_buf[3] = (scsi_buf[3] & 0xf0) | 2;

			break;

		case MODE_SENSE:
		case MODE_SELECT:

			/*
			 * TODO. Probably need to change mode sense/select to 10 byte versions
			 */

		default:
			break;
	}
	return;
}

/*
 * This function deals with status writes from the SBP-2 device
 */
static int sbp2_handle_status_write(struct hpsb_host *host, int nodeid, quadlet_t *data,
				    u64 addr, unsigned int length)
{
	struct sbp2scsi_host_info *hi = NULL;
	struct scsi_id_instance_data *scsi_id = NULL;
	int i;
	unsigned long flags;
	Scsi_Cmnd *SCpnt = NULL;

	SBP2_DEBUG("sbp2: sbp2_handle_status_write");

	if (!host) {
		SBP2_ERR("sbp2: host is NULL - this is bad!");
		return(RCODE_ADDRESS_ERROR);
	}

	spin_lock_irqsave(&sbp2_host_info_lock, flags);
	hi = sbp2_find_host_info(host);
	spin_unlock_irqrestore(&sbp2_host_info_lock, flags);

	if (!hi) {
		SBP2_ERR("sbp2: host info is NULL - this is bad!");
		return(RCODE_ADDRESS_ERROR);
	}

	spin_lock_irqsave(&hi->sbp2_command_lock, flags);

	/*
	 * Find our scsi_id structure
	 */
	for (i=0; i<SBP2SCSI_MAX_SCSI_IDS; i++) {
		if (hi->scsi_id[i]) {
			if (hi->scsi_id[i]->node_id == (nodeid & NODE_MASK)) {
				scsi_id = hi->scsi_id[i];
				SBP2_DEBUG("sbp2: SBP-2 status write from node %x", scsi_id->node_id);
				break;
			}
		}
	}

	if (!scsi_id) {
		SBP2_ERR("sbp2: scsi_id is NULL - device is gone?");
		spin_unlock_irqrestore(&hi->sbp2_command_lock, flags);
		return(RCODE_ADDRESS_ERROR);
	}

	/*
	 * Put response into scsi_id status fifo... 
	 */
	memcpy(&scsi_id->status_block, data, length);

	/*
	 * Byte swap first two quadlets (8 bytes) of status for processing
	 */
	sbp2util_be32_to_cpu_buffer(&scsi_id->status_block, 8);

	/*
	 * Handle command ORB status here if necessary. First, need to match status with command.
	 */
	if (bus_to_virt(scsi_id->status_block.ORB_offset_lo) == &scsi_id->command_orb) {

		SBP2_DEBUG("sbp2: Found status for command ORB");

		/*
		 * Matched status with command, now grab scsi command pointers and check status
		 */
		SCpnt = scsi_id->Current_SCpnt;

		if (SCpnt) {

			/*
			 * Handle check conditions
			 */
			if (scsi_id->status_block.sbp_status) {

				SBP2_DEBUG("sbp2: CHECK CONDITION");

				/*
				 * Translate SBP-2 status to SCSI sense data
				 */
				sbp2_status_to_sense_data((unchar *)&scsi_id->status_block, SCpnt->sense_buffer);

				/*
				 * Initiate a fetch agent reset. 
				 */
				sbp2_agent_reset(hi, scsi_id, SBP2_SEND_NO_WAIT);

			}

			/*
			 * Complete the SCSI command
			 */
			SBP2_DEBUG("sbp2: Completing SCSI command");
			if (scsi_id->status_block.sbp_status) {
				sbp2scsi_complete_command(hi, scsi_id, ((unchar *)&scsi_id->status_block)[8] & 0x3f);
			} else {
				sbp2scsi_complete_command(hi, scsi_id, SBP2_SCSI_STATUS_GOOD);
			}

		}

	}

	spin_unlock_irqrestore(&hi->sbp2_command_lock, flags);
	wake_up(&scsi_id->sbp2_login_wait);
	return(RCODE_COMPLETE);
}


/**************************************
 * SCSI interface related section
 **************************************/

/*
 * This routine is used to deal with the older sbp2scsi_command routine
 * that expects to be completed synchronously. Note: Not used, only here 
 * for completeness.
 */
static void sbp2scsi_internal_done (Scsi_Cmnd *SCpnt) 
{
	SCpnt->SCp.Status++;
}

/*
 * This routine handles the older scsi method of doing I/O (synchronous
 * I/O). This routine is called directly by the SCSI stack. Note: Not used,
 * only here for completeness.
 */
static int sbp2scsi_command (Scsi_Cmnd *SCpnt) 
{

	/*
	 * Queue up the command
	 */
	SCpnt->SCp.Status = 0;
	sbp2scsi_queuecommand (SCpnt, sbp2scsi_internal_done);

	/*
	 * Spin on completion
	 */
	while (!SCpnt->SCp.Status);

	/*
	 * Return result
	 */
	return(SCpnt->result);
}

/*
 * This routine is the main request entry routine for doing I/O. It is 
 * called from the scsi stack directly.
 */
static int sbp2scsi_queuecommand (Scsi_Cmnd *SCpnt, void (*done)(Scsi_Cmnd *)) 
{
	struct sbp2scsi_host_info *hi = NULL;
	struct scsi_id_instance_data *scsi_id = NULL;
	unsigned long flags;

	SBP2_DEBUG("sbp2: sbp2scsi_queuecommand");

	/*
	 * Pull our host info and scsi id instance data from the scsi command
	 */
	hi = (struct sbp2scsi_host_info *) SCpnt->host->hostdata[0];

	if (!hi) {
		SBP2_ERR("sbp2: sbp2scsi_host_info is NULL - this is bad!");
		SCpnt->result = DID_NO_CONNECT << 16;
		done (SCpnt);
		return(0);
	}

	scsi_id = hi->scsi_id[SCpnt->target];

	/*
	 * If scsi_id is null, it means there is no device in this slot, so we should return 
	 * selection timeout.
	 */
	if (!scsi_id) {
		SCpnt->result = DID_NO_CONNECT << 16;
		done (SCpnt);
		return(0);
	}

	/*
	 * Until we handle multiple luns, just return selection time-out to any IO directed at non-zero LUNs
	 */
	if (SCpnt->lun) {
		SCpnt->result = DID_NO_CONNECT << 16;
		done (SCpnt);
		return(0);
	}

	/*
	 * Sanity! Should not normally occur. 
	 */
	if (scsi_id->Current_SCpnt || scsi_id->Current_done) {
		SBP2_ERR("sbp2: Overlapped SCSI command to same scsi id - this is bad!");
		SCpnt->result = DID_NO_CONNECT << 16;
		done (SCpnt);
		return(0);
	}

	/*
	 * Save off our current scsi command and done routine
	 */
	scsi_id->Current_SCpnt = SCpnt;
	scsi_id->Current_done = done;

	/*
	 * Check for request sense command, and handle it here (autorequest sense)
	 */
	if (SCpnt->cmnd[0] == REQUEST_SENSE) {
		SBP2_ERR("sbp2: REQUEST_SENSE");
		memcpy(SCpnt->request_buffer, SCpnt->sense_buffer, SCpnt->request_bufflen);
		memset(SCpnt->sense_buffer, 0, sizeof(SCpnt->sense_buffer));
		sbp2scsi_complete_command(hi, scsi_id, SBP2_SCSI_STATUS_GOOD);
		return(0);
	}

	/*
	 * Check to see if there is a bus reset in progress. If so, we'll execute the command
	 * later after the bus reset is complete.
	 */
	if (hi->bus_reset_in_progress) {
		SBP2_ERR("sbp2: Bus reset in progress - delaying request");
		return(0);
	}

	/*
	 * Try and send our SCSI command
	 */
	spin_lock_irqsave(&hi->sbp2_command_lock, flags);
	if (sbp2_send_command(hi, scsi_id)) {
		SBP2_ERR("sbp2: Error sending SCSI command");
		sbp2scsi_complete_command(hi, scsi_id, SBP2_SCSI_STATUS_SELECTION_TIMEOUT);
	}
	spin_unlock_irqrestore(&hi->sbp2_command_lock, flags);

	return(0);
}

/*
 * This function is called in order to complete a regular SBP-2 command. 
 */
static void sbp2scsi_complete_command(struct sbp2scsi_host_info *hi, struct scsi_id_instance_data *scsi_id, u32 scsi_status)
{
	void (*done)(Scsi_Cmnd *) = scsi_id->Current_done;
	Scsi_Cmnd *SCpnt = scsi_id->Current_SCpnt;

	SBP2_DEBUG("sbp2: sbp2scsi_complete_command");

	/*
	 * Sanity
	 */
	if (!SCpnt) {
		SBP2_ERR("sbp2: SCpnt is NULL");
		return;
	}

	/*
	 * If a bus reset is in progress and there was an error, don't complete the command,
	 * just let it get retried at the end of the bus reset.
	 */
	if ((hi->bus_reset_in_progress) && (scsi_status != SBP2_SCSI_STATUS_GOOD)) {
		SBP2_ERR("sbp2: Bus reset in progress - retry command later");
		return;
	}

	/*
	 * Retry the command if the device returns unit attention... it could have happened because
	 * of a 1394 bus reset or hot-plug...
	 */
#ifdef SBP2_DO_RETRY_ON_UNIT_ATTENTION
	if ((scsi_status == SBP2_SCSI_STATUS_CHECK_CONDITION) && (SCpnt->sense_buffer[2] == UNIT_ATTENTION)) {
		SBP2_ERR("sbp2: UNIT ATTENTION - retry command");
		if (sbp2_send_command(hi, scsi_id)) {
			SBP2_ERR("sbp2: Error sending SCSI command");
		} else {
			return;
		}
	}
#endif

	/*
	 * Null out current scsi command pointers
	 */
	scsi_id->Current_SCpnt = NULL;
	scsi_id->Current_done = NULL;

	/*
	 * Switch on scsi status
	 */
	switch (scsi_status) {
		case SBP2_SCSI_STATUS_GOOD:
			SCpnt->result = DID_OK;
			break;

		case SBP2_SCSI_STATUS_BUSY:
			SBP2_ERR("sbp2: SBP2_SCSI_STATUS_BUSY");
			SCpnt->result = DID_BUS_BUSY << 16;
			break;

		case SBP2_SCSI_STATUS_CHECK_CONDITION:
			SBP2_ERR("sbp2: SBP2_SCSI_STATUS_CHECK_CONDITION");
			SCpnt->result = CHECK_CONDITION << 1;

			/*
			 * Debug stuff
			 */
			print_sense("bh", SCpnt);

			break;

		case SBP2_SCSI_STATUS_SELECTION_TIMEOUT:
			SBP2_ERR("sbp2: SBP2_SCSI_STATUS_SELECTION_TIMEOUT");
			SCpnt->result = DID_NO_CONNECT << 16;
			break;

		case SBP2_SCSI_STATUS_CONDITION_MET:
		case SBP2_SCSI_STATUS_RESERVATION_CONFLICT:
		case SBP2_SCSI_STATUS_COMMAND_TERMINATED:
			SBP2_ERR("sbp2: Bad SCSI status = %x", scsi_status);
			SCpnt->result = DID_ERROR << 16;
			break;

		default:
			SBP2_ERR("sbp2: Unsupported SCSI status = %x", scsi_status);
			SCpnt->result = DID_ERROR << 16;
	}

	/*
	 * Take care of any sbp2 response data mucking here (RBC stuff, etc.)
	 */
	if (SCpnt->result == DID_OK) {
		sbp2_check_sbp2_response(SCpnt);
	}

	/*
	 * Tell scsi stack that we're done with this command
	 */
	done (SCpnt);

	return;
}

/*
 * Called by scsi stack when something has really gone wrong (enhanced version).
 * Usually called when a command has timed-out for some reason. 
 */
static int sbp2scsi_eh_abort (Scsi_Cmnd *SCpnt) {
	sbp2scsi_abort(SCpnt);
	return(SUCCESS);       
}

/*
 * Called by scsi stack when something has really gone wrong.
 * Usually called when a command has timed-out for some reason. 
 */
static int sbp2scsi_abort (Scsi_Cmnd *SCpnt) 
{
	struct sbp2scsi_host_info *hi = NULL;
	struct scsi_id_instance_data *scsi_id = NULL;
	void (*done)(Scsi_Cmnd *) = SCpnt->done;
	unsigned long flags;

	SBP2_ERR("sbp2: aborting sbp2 command");

	/*
	 * Pull our host info and scsi id instance data from the scsi command
	 */
	hi = (struct sbp2scsi_host_info *) SCpnt->host->hostdata[0];

	if (!hi) {
		SBP2_ERR("sbp2: sbp2scsi_host_info is NULL - this is bad!");
		SCpnt->result = DID_ABORT << 16;
		return(SCSI_ABORT_SNOOZE);
	}

	scsi_id = hi->scsi_id[SCpnt->target];

	if (!scsi_id) {
		done = SCpnt->done;
		SCpnt->result = DID_ABORT << 16;
		done (SCpnt);
		return(SCSI_ABORT_SUCCESS);
	}

	spin_lock_irqsave(&hi->sbp2_command_lock, flags);

	/*
	 * Initiate a fetch agent reset. 
	 */
	sbp2_agent_reset(hi, scsi_id, SBP2_SEND_NO_WAIT);

	/*
	 * Null out current command and complete the request. If we are in the middle of a reset, complete
	 * with busy status.
	 */
	if (scsi_id->Current_done) {
		done = scsi_id->Current_done;
		scsi_id->Current_SCpnt = NULL;
		scsi_id->Current_done = NULL;
		SCpnt->result = DID_ABORT << 16;
		done (SCpnt);
	}

	spin_unlock_irqrestore(&hi->sbp2_command_lock, flags);
	return(SCSI_ABORT_SUCCESS);
}

/*
 * Called by scsi stack when something has really gone wrong. 
 */
static int sbp2scsi_eh_reset (Scsi_Cmnd *SCpnt) {
	SBP2_ERR("sbp2: reset requested");
	return(SUCCESS);       
}

/*
 * Called by scsi stack when something has really gone wrong.
 */
static int sbp2scsi_reset (Scsi_Cmnd *SCpnt, unsigned int reset_flags) 
{
	SBP2_ERR("sbp2: reset requested");
	return(SCSI_RESET_SUCCESS);
}

/*
 * Called by scsi stack to get bios parameters (used by fdisk, and at boot).
 */
static int sbp2scsi_biosparam (Scsi_Disk *disk, kdev_t dev, int geom[]) 
{
	int heads, sectors, cylinders;

	SBP2_ERR("sbp2: request for bios parameters");

	heads = 64;
	sectors = 32;
	cylinders = disk->capacity / (heads * sectors);

	if (cylinders > 1024) {
		heads = 255;
		sectors = 63;
		cylinders = disk->capacity / (heads * sectors);
	}

	geom[0] = heads;
	geom[1] = sectors;
	geom[2] = cylinders;

	return(0);
}

/*
 * This routine is called at setup (init) and does nothing. Not used here.   =)
 */
void sbp2scsi_setup( char *str, int *ints) 
{
	SBP2_DEBUG("sbp2: sbp2scsi_setup");
	return;
}

/*
 * This is our detection routine, and is where we init everything.
 */
static int sbp2scsi_detect (Scsi_Host_Template *tpnt) 
{
	SBP2_DEBUG("sbp2: sbp2scsi_detect");

	printk("SBP-2 protocol driver for IEEE-1394\n");

	global_scsi_tpnt = tpnt;

#if LINUX_VERSION_CODE < 0x20300
	global_scsi_tpnt->proc_dir = &proc_scsi_sbp2scsi;
#else
	global_scsi_tpnt->proc_name = "sbp2";
#endif

	sbp2_init();

	if (!sbp2_host_count) {
		printk("Please load the lower level IEEE-1394 driver (e.g. ohci1394) before sbp2_1394...\n");
		if (sbp2_hl_handle) {
			hpsb_unregister_highlevel(sbp2_hl_handle);
			sbp2_hl_handle = NULL;
		}
	}

	/*
	 * Since we are returning this count, it means that sbp2 must be loaded "after" the 
	 * host adapter module... 	 
	 */
	return(sbp2_host_count);
}

/*
 * This function is called from sbp2_add_host, and is where we register our scsi host
 */
static void sbp2scsi_register_scsi_host(struct sbp2scsi_host_info *hi)
{
	struct Scsi_Host *shpnt = NULL;

	SBP2_DEBUG("sbp2: sbp2scsi_register_scsi_host");
	SBP2_DEBUG("sbp2: sbp2scsi_host_info = %x", (unsigned int) hi);

	/*
	 * Let's register with the scsi stack
	 */
	if (global_scsi_tpnt) {

		shpnt = scsi_register (global_scsi_tpnt, sizeof(void *));

		/*
		 * If successful, save off a context (to be used when SCSI commands are received)
		 */
		if (shpnt) {
			shpnt->hostdata[0] = (unsigned long)hi;
		}
	}

	return;
}

/*
 * Called when our module is released
 */
static int sbp2scsi_release(struct Scsi_Host *host)
{
	SBP2_DEBUG("sbp2: sbp2scsi_release");
	sbp2_cleanup();
	return(0);
}

/*
 * Called for contents of procfs
 */
static const char *sbp2scsi_info (struct Scsi_Host *host)
{
	return "IEEE-1394 SBP-2 protocol driver";
}

/*
 * Module related section
 */

#ifdef MODULE

MODULE_AUTHOR("James Goodwin <jamesg@filanet.com>");
MODULE_DESCRIPTION("IEEE-1394 SBP-2 protocol driver");
MODULE_SUPPORTED_DEVICE("sbp2_1394");

/*
 * SCSI host template
 */
Scsi_Host_Template driver_template = SBP2SCSI;
#include "../scsi/scsi_module.c"

#endif
