/*
 *  linux/fs/ext3/file.c
 *
 * Copyright (C) 1992, 1993, 1994, 1995
 * Remy Card (card@masi.ibp.fr)
 * Laboratoire MASI - Institut Blaise Pascal
 * Universite Pierre et Marie Curie (Paris VI)
 *
 *  from
 *
 *  linux/fs/minix/file.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *
 *  ext3 fs regular file handling primitives
 *
 *  64-bit file support on 64-bit platforms by Jakub Jelinek
 * 	(jj@sunsite.ms.mff.cuni.cz)
 */

#include <asm/uaccess.h>
#include <asm/system.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/ext3_fs.h>
#include <linux/ext3_jfs.h>
#include <linux/fcntl.h>
#include <linux/stat.h>
#include <linux/locks.h>
#include <linux/mm.h>
#include <linux/pagemap.h>

#define	NBUF	32

#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

static long long ext3_file_lseek(struct file *, long long, int);
static ssize_t ext3_file_write (struct file *, const char *, size_t, loff_t *);
static int ext3_release_file (struct inode *, struct file *);

#define EXT3_MAX_SIZE(bits)							\
	(((EXT3_NDIR_BLOCKS + (1LL << (bits - 2)) + 				\
	   (1LL << (bits - 2)) * (1LL << (bits - 2)) + 				\
	   (1LL << (bits - 2)) * (1LL << (bits - 2)) * (1LL << (bits - 2))) * 	\
	  (1LL << bits)) - 1)

long long ext3_max_sizes[] = {
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
EXT3_MAX_SIZE(10), EXT3_MAX_SIZE(11), EXT3_MAX_SIZE(12), EXT3_MAX_SIZE(13)
};

/*
 * We have mostly NULL's here: the current defaults are ok for
 * the ext3 filesystem.
 */
static struct file_operations ext3_file_operations = {
	ext3_file_lseek,	/* lseek */
	generic_file_read,	/* read */
	ext3_file_write,	/* write */
	NULL,			/* readdir - bad */
	NULL,			/* poll - default */
	ext3_ioctl,		/* ioctl */
	generic_file_mmap,	/* mmap */
	NULL,			/* no special open is needed */
	NULL,			/* flush */
	ext3_release_file,	/* release */
	ext3_sync_file,		/* fsync */
	NULL,			/* fasync */
	NULL,			/* check_media_change */
	NULL			/* revalidate */
};

struct inode_operations ext3_file_inode_operations = {
	&ext3_file_operations,/* default file operations */
	NULL,			/* create */
	NULL,			/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	generic_readpage,	/* readpage */
	NULL,			/* writepage */
	ext3_bmap,		/* bmap */
	ext3_truncate,		/* truncate */
	ext3_permission,	/* permission */
	NULL			/* smap */
};


/*
 * Make sure the offset never goes beyond the 32-bit mark..
 */
static long long ext3_file_lseek(
	struct file *file,
	long long offset,
	int origin)
{
	struct inode *inode = file->f_dentry->d_inode;

	switch (origin) {
		case 2:
			offset += inode->i_size;
			break;
		case 1:
			offset += file->f_pos;
	}
	if (offset < 0 ||
	    offset > ext3_max_sizes[EXT3_BLOCK_SIZE_BITS(inode->i_sb)])
		return -EINVAL;
	if (offset != file->f_pos) {
		file->f_pos = offset;
		file->f_reada = 0;
		file->f_version = ++global_event;
	}
	return offset;
}

inline void ext3_remove_suid(struct inode *inode)
{
	unsigned int mode;

	/* set S_IGID if S_IXGRP is set, and always set S_ISUID */
	mode = (inode->i_mode & S_IXGRP)*(S_ISGID/S_IXGRP) | S_ISUID;

	/* was any of the uid bits set? */
	mode &= inode->i_mode;
	if (mode && !capable(CAP_FSETID)) {
		inode->i_mode &= ~mode;
		mark_inode_dirty(inode);
	}
}

static ssize_t ext3_file_write (struct file * filp, const char * buf,
				size_t count, loff_t *ppos)
{
	struct inode * inode = filp->f_dentry->d_inode;
	loff_t pos;
	long block, needed;
	int offset;
	size_t written, written_transaction, c;
	struct buffer_head * bh;
	struct super_block * sb;
	int err;
	int write_error, new_buffer;
	unsigned long limit;
	handle_t *handle;
	int will_journal_data, ordered_data;
	
	/* POSIX: mtime/ctime may not change for 0 count */
	if (!count)
		return 0;
	/* This makes the bounds-checking arithmetic later on much more
	 * sane. */
	if (((ssize_t) count) < 0)
		return -EINVAL;
	
	jfs_debug(4, "Write for %u at %ld to inode %p\n",
		  count, (long) *ppos, inode);
	write_error = 0;
	if (!inode) {
		printk("ext3_file_write: inode = NULL\n");
		return -EINVAL;
	}
	sb = inode->i_sb;
	if (sb->s_flags & MS_RDONLY)
		/*
		 * This fs has been automatically remounted ro because of errors
		 */
		return -ENOSPC;

	if (!S_ISREG(inode->i_mode)) {
		ext3_warning (sb, "ext3_file_write", "mode = %07o",
			      inode->i_mode);
		return -EINVAL;
	}
	ext3_remove_suid(inode);

	if (filp->f_flags & O_APPEND)
		pos = inode->i_size;
	else {
		pos = *ppos;
		if (pos != *ppos)
			return -EINVAL;
	}

	/* Check for overflow.. */

	/* L-F-S spec 2.2.1.27: */
	if (!(filp->f_flags & O_LARGEFILE)) {
		if (pos >= 0x7fffffffULL) /* pos@2G forbidden */
			return -EFBIG;

		if (pos + count > 0x7fffffffULL)
			/* Write only until end of allowed region */
			count = 0x7fffffffULL - pos;
	}

	{
		loff_t max = ext3_max_sizes[EXT3_BLOCK_SIZE_BITS(sb)];

		if (pos >= max)
			return -EFBIG;

		if (pos + count > max) {
			count = max - pos;
			if (!count)
				return -EFBIG;
		}
	}

	/* From SUS: We must generate a SIGXFSZ for file size overflow
	 * only if no bytes were actually written to the file. --sct */

	limit = current->rlim[RLIMIT_FSIZE].rlim_cur;
	if (limit != RLIM_INFINITY) {
		if (pos >= limit) {
			send_sig(SIGXFSZ, current, 0);
			return -EFBIG;
		}
		if (pos+count > limit)
			count = limit - pos;
	}

	/* will_journal_data must not just check the journaling mode:
	 * specific files can be marked for data-journaling even with
	 * ordered or writeback journaling (eg. the quota files).
	 *
	 * ordered_data is set if we are using ordered data for writes,
	 * OR if O_SYNC is specified.  In the latter case, we'll
	 * automatically trigger a data writeback as a result of the
	 * commit-on-journal_stop. */
	will_journal_data = ext3_should_journal_data(inode);
	ordered_data = ((test_opt(inode->i_sb, DATA_FLAGS) == 
			 EXT3_MOUNT_ORDERED_DATA) || 
			(filp->f_flags & O_SYNC));		
	
	jfs_debug(6, "will_journal_data is %d, ordered_data is %d\n", 
		  will_journal_data, ordered_data);
	
	/* How large a transaction might we need?  We can always
	 * underestimate and grow later for really large writes */
	if (will_journal_data) {
		needed = (count >> EXT3_BLOCK_SIZE_BITS(sb)) + 1;
		if (needed > EXT3_MAX_TRANS_DATA)
			needed = EXT3_MAX_TRANS_DATA;
	} else
		needed = 0;

	block = pos >> EXT3_BLOCK_SIZE_BITS(sb);
	offset = pos & (sb->s_blocksize - 1);
	c = sb->s_blocksize - offset;

	handle = ext3_journal_start(inode, EXT3_DATA_TRANS_BLOCKS + needed);

	if (((pos + count) >> 31) &&
	    (!EXT3_HAS_RO_COMPAT_FEATURE(sb,EXT3_FEATURE_RO_COMPAT_LARGE_FILE)||
	     EXT3_SB(sb)->s_es->s_rev_level == cpu_to_le32(EXT3_GOOD_OLD_REV))){
		/* If this is the first large file, add a flag to superblock */
		journal_get_write_access(handle, sb->u.ext3_sb.s_sbh);
		ext3_update_fs_rev(sb);
		EXT3_SET_RO_COMPAT_FEATURE(sb,
					   EXT3_FEATURE_RO_COMPAT_LARGE_FILE);
		journal_dirty_metadata(handle, sb->u.ext3_sb.s_sbh);
	}

	written = written_transaction = 0;
	do {
		/* Each time round the loop, work out whether we can
		 * continue as part of the current transaction or
		 * whether we need to start another one.  This obviously
		 * matters only for large writes. */
		if (handle->h_buffer_credits < EXT3_RESERVE_TRANS_BLOCKS &&
		    written_transaction >= (EXT3_MAX_TRANS_DATA/2)) {
			needed = (count >> EXT3_BLOCK_SIZE_BITS(sb)) + 1;
			if (needed > EXT3_MAX_TRANS_DATA)
				needed = EXT3_MAX_TRANS_DATA;
			if (journal_extend(handle, needed)) {
				/* Couldn't extend: OK, commit the current
				 * transaction and start a new one. */
				if (pos > inode->u.ext3_i.i_disksize)
					inode->u.ext3_i.i_disksize = pos;
				inode->i_ctime = inode->i_mtime = CURRENT_TIME;
				ext3_mark_inode_dirty(handle, inode);
				err = ext3_journal_stop(handle, inode);
				if (err) {
					written = err;
					break;
				}
				
				handle = ext3_journal_start
				    (inode, EXT3_DATA_TRANS_BLOCKS + needed);
				written_transaction = 0;
			}
		}

		bh = ext3_getblk (handle, inode, block, 1, &err);
		if (!bh) {
			if (!written)
				written = err;
			break;
		}
		if (c > count)
			c = count;

		/* Tricky: what happens if we are writing the complete
		 * contents of a block which is not currently
		 * initialised?  We have to obey the same
		 * synchronisation rules as the IO code, to prevent some
		 * other process from stomping on the buffer contents by
		 * refreshing them from disk while we are setting up the
		 * buffer.  The copy_from_user() can page fault, after
		 * all.  We also have to throw away partially successful
		 * copy_from_users to such buffers, since we can't trust
		 * the rest of the buffer_head in that case.  --sct */

		new_buffer = (!buffer_uptodate(bh) && !buffer_locked(bh) &&
			      c == sb->s_blocksize);

		if (new_buffer) {
			set_bit(BH_Lock, &bh->b_state);
			if (will_journal_data)
				journal_get_create_access(handle, bh);
			c -= copy_from_user (bh->b_data + offset, buf, c);
			if (c != sb->s_blocksize) {
				/* On getting an EFAULT mid-copy, we end
				 * up throwing away the whole block.
				 * Too bad.  */
				c = 0;
				unlock_buffer(bh);
				if (will_journal_data)
					journal_release_buffer(handle, bh);
				brelse(bh);
				if (!written)
					written = -EFAULT;
				break;
			}
			mark_buffer_uptodate(bh, 1);
			unlock_buffer(bh);
		} else {
			if (!buffer_uptodate(bh)) {
				ll_rw_block (READ, 1, &bh);
				wait_on_buffer (bh);
				if (!buffer_uptodate(bh)) {
					brelse (bh);
					if (!written)
						written = -EIO;
					break;
				}
			}
			if (will_journal_data)
				journal_get_write_access(handle, bh);
			c -= copy_from_user (bh->b_data + offset, buf, c);
		}
		if (!c) {
			journal_release_buffer(handle, bh);
			brelse(bh);
			if (!written)
				written = -EFAULT;
			break;
		}

		if (will_journal_data)
			journal_dirty_metadata(handle, bh);
		else if (ordered_data)
			journal_dirty_data(handle, bh);
		else
			mark_buffer_dirty(bh, 0);
		
		update_vm_cache(inode, pos, bh->b_data + offset, c);
		pos += c;
		written += c;
		written_transaction += c;
		buf += c;
		count -= c;
		brelse(bh);

		if (write_error)
			break;

		block++;
		offset = 0;
		c = sb->s_blocksize;
	} while (count);

	/*
	 * If a file has been opened in synchronous mode, we have to ensure
	 * that meta-data will also be written synchronously.  
	 */
	if (filp->f_flags & O_SYNC)
		handle->h_sync = 1;

	if (pos > inode->i_size) {
		inode->i_size = pos;
		inode->u.ext3_i.i_disksize = pos;
	}
	
	inode->i_ctime = inode->i_mtime = CURRENT_TIME;
	*ppos = pos;
	ext3_mark_inode_dirty(handle, inode);
	ext3_journal_stop(handle, inode);
	return written;
}

/*
 * Called when an inode is released. Note that this is different
 * from ext3_file_open: open gets called at every open, but release
 * gets called only when /all/ the files are closed.
 */
static int ext3_release_file (struct inode * inode, struct file * filp)
{
	if (filp->f_mode & FMODE_WRITE)
		ext3_discard_prealloc (inode);
	return 0;
}

