/*
 *	linux/mm/filemap.c
 *
 * Copyright (C) 1994, 1995  Linus Torvalds
 * 
 * KAIO modifications:
 * 	Copyright 1999, Silicon Graphics, Inc.
 *
 * 	Written October 1999 by Rajagopal Ananthanarayanan (ananth)
 *                      at Silicon Graphics, Inc.
 * 	With modifications suggested by Doug Rady at Silicon Graphics, Inc.
 */

/*
 * This file handles the generic file mmap semantics used by
 * most "normal" filesystems (but you don't /have/ to use this:
 * the NFS filesystem used to do this differently, for example)
 */
#include <linux/malloc.h>
#include <linux/shm.h>
#include <linux/mman.h>
#include <linux/locks.h>
#include <linux/pagemap.h>
#include <linux/swap.h>
#include <linux/smp_lock.h>
#include <linux/blkdev.h>
#include <linux/file.h>
#include <linux/swapctl.h>
#include <linux/init.h>
#include <linux/bigmem.h>

#include <asm/pgtable.h>
#include <asm/uaccess.h>

#if defined(CONFIG_AIO)
#include <linux/aio.h>

#ifdef AIO_DEBUG
#define	assert(expression) { \
	if (!(expression)) { \
		(void)printk(KERN_ERR \
			"assertion \"%s\" failed: file \"%s\", line %d\n", \
			#expression, \
			__FILE__, __LINE__); \
	} \
}
#else
#define assert(expresssion)
#endif

#define pageof(x)	(x >> PAGE_SHIFT)
#endif /* CONFIG_AIO */

/*
 * Shared mappings implemented 30.11.1994. It's not fully working yet,
 * though.
 *
 * Shared mappings now work. 15.8.1995  Bruno.
 */

unsigned long page_cache_size = 0;
unsigned int page_hash_bits, page_hash_mask;
struct page **page_hash_table;

static inline int sync_page(struct page *page)
{
	struct inode *inode = page->inode;

	if (inode && inode->i_op && inode->i_op->sync_page)
		return inode->i_op->sync_page(page);
	run_task_queue(&tq_disk);
	return 0;
}

/*
 * Invalidate the pages of an inode, removing all pages that aren't
 * locked down (those are sure to be up-to-date anyway, so we shouldn't
 * invalidate them).
 */
void invalidate_inode_pages(struct inode * inode)
{
	struct page ** p;
	struct page * page;

	p = &inode->i_pages;
	while ((page = *p) != NULL) {
		if (PageLocked(page)) {
			p = &page->next;
			continue;
		}
		inode->i_nrpages--;
		if ((*p = page->next) != NULL)
			(*p)->prev = page->prev;
		page->next = NULL;
		page->prev = NULL;
		remove_page_from_hash_queue(page);
		page->inode = NULL;
		page_cache_release(page);
		continue;
	}
}

/*
 * Truncate the page cache at a set offset, removing the pages
 * that are beyond that offset (and zeroing out partial pages).
 */
void truncate_inode_pages(struct inode * inode, loff_t start)
{
	struct page ** p;
	struct page * page;

repeat:
	p = &inode->i_pages;
	while ((page = *p) != NULL) {
		loff_t loffset = pgoff2loff(page->index);

		/* page wholly truncated - free it */
		if (loffset >= start) {
			if (PageLocked(page)) {
				wait_on_page(page);
				goto repeat;
			}
			inode->i_nrpages--;
			if ((*p = page->next) != NULL)
				(*p)->prev = page->prev;
			page->next = NULL;
			page->prev = NULL;
			remove_page_from_hash_queue(page);
			page->inode = NULL;
			page_cache_release(page);
			continue;
		}
		p = &page->next;
		loffset = start - loffset;
		/* partial truncate, clear end of page */
		if (loffset < PAGE_CACHE_SIZE) {
			unsigned int  offset  = loffset; /* truncate ok */
			unsigned long address = page_address(page);
			memset((void *) (offset + address), 0, PAGE_CACHE_SIZE - offset);
			flush_page_to_ram(address);
		}
	}
}

/*
 * Remove a page from the page cache and free it.
 */
void remove_inode_page(struct page *page)
{
	remove_page_from_hash_queue(page);
	remove_page_from_inode_queue(page);
	page_cache_release(page);
}

int shrink_mmap(int priority, int gfp_mask)
{
	static unsigned long clock = 0;
#ifndef CONFIG_BIGMEM
	unsigned long limit = num_physpages;
#else
	unsigned long limit = bigmem_mapnr;
#endif
	struct page * page;
	int count;

	/* Make sure we scan all pages twice at priority 0. */
	count = limit / priority;

 refresh_clock:
	page = mem_map + clock;
	do {
		int referenced;

		if (current->need_resched) {
			current->state = TASK_RUNNING;
			schedule();
			goto refresh_clock;
		}
		
		/* This works even in the presence of PageSkip because
		 * the first two entries at the beginning of a hole will
		 * be marked, not just the first.
		 */
		page++;
		clock++;
#ifndef CONFIG_BIGMEM
		if (clock >= max_mapnr) {
#else
		if (clock >= bigmem_mapnr) {
#endif
			clock = 0;
			page = mem_map;
		}
		if (PageSkip(page)) {
			/* next_hash is overloaded for PageSkip */
			page = page->next_hash;
			clock = page - mem_map;
		}
		
		count--;

		/* We can't free pages unless there's just one user */
		if (atomic_read(&page->count) != 1)
			continue;

		referenced = test_and_clear_bit(PG_referenced, &page->flags);

		if (PageLocked(page))
			continue;

		if (!(gfp_mask & __GFP_BIGMEM) && PageBIGMEM(page)) {
			count++;
			continue;
		}

		if ((gfp_mask & __GFP_DMA) && !PageDMA(page)) {
			count++;
			continue;
		}

		/*
		 * Is it a page swap page? If so, we want to
		 * drop it if it is no longer used, even if it
		 * were to be marked referenced..
		 */
		if (PageSwapCache(page)) {
			if (referenced &&
			    swap_count(pgoff2ulong(page->index)) != 1)
				continue;
			delete_from_swap_cache(page);
			return 1;
		}	

		if (referenced)
			continue;

		/* Is it a buffer page? */
		if (page->buffers) {
			if (buffer_under_min())
				continue;
			/*
			 * We can sleep if we need to do some write
			 * throttling.
			 */

			if (!try_to_free_buffers(page, gfp_mask))
				goto refresh_clock;
			return 1;
		}

		/* is it a page-cache page? */
		if (page->inode) {
			if (pgcache_under_min())
				continue;
			remove_inode_page(page);
			return 1;
		}
	} while (count > 0);
	return 0;
}

/*
 * Update a page cache copy, when we're doing a "write()" system call
 * See also "update_vm_cache()".
 *
 * This function is conditional in that it checks whether the original
 * source of the data is the same as the ultimate destination, and
 * aborts the update if so.  
 *
 * The "source_address" is the virtual address of the original location
 * of the data we are injecting.  For writes from user mode, it is the
 * user VA.  However, for filemap_sync writes, "source_address", it is
 * the page cache address.  In both cases, "buf" points to the copy we
 * have already made in kernel space and we use that pointer for the
 * transfer.  source_address just allows us to detect an update_vm_cache
 * which is being sourced from the copy of the data already in the page
 * cache.  
 * 
 * This prevents munmap() and msync() from stomping all over shared
 * memory maps.  --sct
 */

void update_vm_cache_conditional(struct inode * inode, loff_t pos, const char * buf, int count, unsigned long source_address)
{
	unsigned long offset, len;
	pgoff_t pgoff = loff2pgoff(pos);

	offset = ((unsigned long)pos & ~PAGE_CACHE_MASK);
	pos = pos & PAGE_CACHE_MASK;
	len = PAGE_CACHE_SIZE - offset;
	do {
		struct page * page;

		if (len > count)
			len = count;
		page = find_page(inode, pgoff);
		if (page) {
			char *dest = (char*) (offset + page_address(page));

			if ((unsigned long)dest != source_address) {
				wait_on_page(page);
				memcpy(dest, buf, len);
				flush_dcache_page(page_address(page));
			}
			page_cache_release(page);
		}
		count -= len;
		buf += len;
		len = PAGE_CACHE_SIZE;
		offset = 0;
		pos += PAGE_CACHE_SIZE;
	} while (count);
}

void update_vm_cache(struct inode * inode, loff_t pos, const char * buf, int count)
{
	update_vm_cache_conditional(inode, pos, buf, count, 0);
}


static inline void add_to_page_cache(struct page * page,
				     struct inode * inode,
				     pgoff_t pgoff,
				     struct page **hash)
{
	atomic_inc(&page->count);
	page->flags = page->flags & ~((1 << PG_uptodate) | (1 << PG_error) | (1 << PG_referenced));
	page->index = pgoff;
	add_page_to_inode_queue(inode, page);
	__add_page_to_hash_queue(page, hash);
}

/*
 * Try to read ahead in the file. "page_cache" is a potentially free page
 * that we could use for the cache (if it is 0 we can try to create one,
 * this is all overlapped with the IO on the previous page finishing anyway)
 */
static unsigned long try_to_read_ahead(struct file * file,
				       pgoff_t pgoff, unsigned long page_cache)
{
	struct inode *inode = file->f_dentry->d_inode;
	pgoff_t pg_size;

	/* Calculate file size in 'pages' -- if even one byte (according to
	   the 'i_size') exceeds the final page-size block, round up. */
	pg_size = loff2pgoff(inode->i_size+(PAGE_SIZE-1));

	if (!page_cache) {
		page_cache = page_cache_alloc();
		if (!page_cache)
			return 0; /* Can't allocate! */
	}
	/* Ok, we have a page, make sure it is in the page cache */
	if (pgoff2ulong(pgoff) < pgoff2ulong(pg_size)) {
		struct page * page;
		struct page ** hash;
		hash = page_hash(inode, pgoff);
		page = __find_page(inode, pgoff, *hash);
		if (!page) {
			/*
			 * Ok, add the new page to the hash-queues...
			 */
			page = page_cache_entry(page_cache);
			add_to_page_cache(page, inode, pgoff, hash);
			inode->i_op->readpage(file, page);
			page_cache = 0;
		}
		page_cache_release(page);
	}
	return page_cache;
}

/* 
 * Wait for IO to complete on a locked page.
 *
 * This must be called with the caller "holding" the page,
 * ie with increased "page->count" so that the page won't
 * go away during the wait..
 */
void __wait_on_page(struct page *page)
{
	struct task_struct *tsk = current;
	struct wait_queue wait;

	wait.task = tsk;
	add_wait_queue(&page->wait, &wait);
	do {
		set_current_state(TASK_UNINTERRUPTIBLE);
		sync_page(page);
		if (!PageLocked(page))
			break;
		schedule();
	} while (PageLocked(page));
	tsk->state = TASK_RUNNING;
	remove_wait_queue(&page->wait, &wait);
}

#if 0
#define PROFILE_READAHEAD
#define DEBUG_READAHEAD
#endif

/*
 * Read-ahead profiling information
 * --------------------------------
 * Every PROFILE_MAXREADCOUNT, the following information is written 
 * to the syslog:
 *   Percentage of asynchronous read-ahead.
 *   Average of read-ahead fields context value.
 * If DEBUG_READAHEAD is defined, a snapshot of these fields is written 
 * to the syslog.
 */

#ifdef PROFILE_READAHEAD

#define PROFILE_MAXREADCOUNT 1000

static u_long total_reada;
static u_long total_async;
static u_long total_ramax;
static u_long total_ralen;
static u_long total_rawin;

static void profile_readahead(int async, struct file *filp)
{
	unsigned long flags;

	++total_reada;
	if (async)
		++total_async;

	total_ramax	+= filp->f_ramax;
	total_ralen	+= filp->f_ralen;
	total_rawin	+= filp->f_rawin;

	if (total_reada > PROFILE_MAXREADCOUNT) {
		save_flags(flags);
		cli();
		if (!(total_reada > PROFILE_MAXREADCOUNT)) {
			restore_flags(flags);
			return;
		}

		printk("Readahead average:  max=%ld, len=%ld, win=%ld, async=%ld%%\n",
			total_ramax/total_reada,
			total_ralen/total_reada,
			total_rawin/total_reada,
			(total_async*100)/total_reada);
#ifdef DEBUG_READAHEAD
		printk("Readahead snapshot: max=%ld, len=%ld, win=%ld, raend=%ld\n",
			filp->f_ramax, filp->f_ralen, filp->f_rawin, filp->f_raend);
#endif

		total_reada	= 0;
		total_async	= 0;
		total_ramax	= 0;
		total_ralen	= 0;
		total_rawin	= 0;

		restore_flags(flags);
	}
}
#endif  /* defined PROFILE_READAHEAD */

/*
 * Read-ahead context:
 * -------------------
 * The read ahead context fields of the "struct file" are the following:
 * - f_raend : position of the first byte after the last page we tried to
 *             read ahead.
 * - f_ramax : current read-ahead maximum size.
 * - f_ralen : length of the current IO read block we tried to read-ahead.
 * - f_rawin : length of the current read-ahead window.
 *             if last read-ahead was synchronous then
 *                  f_rawin = f_ralen
 *             otherwise (was asynchronous)
 *                  f_rawin = previous value of f_ralen + f_ralen
 *
 * Read-ahead limits:
 * ------------------
 * MIN_READAHEAD   : minimum read-ahead size when read-ahead.
 * MAX_READAHEAD   : maximum read-ahead size when read-ahead.
 *
 * Synchronous read-ahead benefits:
 * --------------------------------
 * Using reasonable IO xfer length from peripheral devices increase system 
 * performances.
 * Reasonable means, in this context, not too large but not too small.
 * The actual maximum value is:
 *	MAX_READAHEAD + PAGE_CACHE_SIZE = 76k is CONFIG_READA_SMALL is undefined
 *      and 32K if defined (4K page size assumed).
 *
 * Asynchronous read-ahead benefits:
 * ---------------------------------
 * Overlapping next read request and user process execution increase system 
 * performance.
 *
 * Read-ahead risks:
 * -----------------
 * We have to guess which further data are needed by the user process.
 * If these data are often not really needed, it's bad for system 
 * performances.
 * However, we know that files are often accessed sequentially by 
 * application programs and it seems that it is possible to have some good 
 * strategy in that guessing.
 * We only try to read-ahead files that seems to be read sequentially.
 *
 * Asynchronous read-ahead risks:
 * ------------------------------
 * In order to maximize overlapping, we must start some asynchronous read 
 * request from the device, as soon as possible.
 * We must be very careful about:
 * - The number of effective pending IO read requests.
 *   ONE seems to be the only reasonable value.
 * - The total memory pool usage for the file access stream.
 *   This maximum memory usage is implicitly 2 IO read chunks:
 *   2*(MAX_READAHEAD + PAGE_CACHE_SIZE) = 156K if CONFIG_READA_SMALL is undefined,
 *   64k if defined (4K page size assumed).
 */

static inline int get_max_readahead(struct inode * inode)
{
	if (!inode->i_dev || !max_readahead[MAJOR(inode->i_dev)])
		return MAX_READAHEAD;
	return max_readahead[MAJOR(inode->i_dev)][MINOR(inode->i_dev)];
}

static inline unsigned long generic_file_readahead(int reada_ok,
	struct file * filp, struct inode * inode,
	loff_t ppos, struct page * page, unsigned long page_cache)
{
	loff_t max_ahead, ahead;
	loff_t raend;
	int max_readahead = get_max_readahead(inode);

	raend = filp->f_raend & PAGE_CACHE_MASK_loff;
	max_ahead = 0;

/*
 * The current page is locked.
 * If the current position is inside the previous read IO request, do not
 * try to reread previously read ahead pages.
 * Otherwise decide or not to read ahead some pages synchronously.
 * If we are not going to read ahead, set the read ahead context for this 
 * page only.
 */
	if (PageLocked(page)) {
		if (!filp->f_ralen || ppos >= raend || ppos + filp->f_ralen < raend) {
			raend = ppos;
			if (raend < inode->i_size)
				max_ahead = filp->f_ramax;
			filp->f_rawin = 0;
			filp->f_ralen = PAGE_CACHE_SIZE;
			if (!max_ahead) {
				filp->f_raend  = ppos + filp->f_ralen;
				filp->f_rawin += filp->f_ralen;
			}
		}
	}
/*
 * The current page is not locked.
 * If we were reading ahead and,
 * if the current max read ahead size is not zero and,
 * if the current position is inside the last read-ahead IO request,
 *   it is the moment to try to read ahead asynchronously.
 * We will later force unplug device in order to force asynchronous read IO.
 */
	else if (reada_ok && filp->f_ramax && raend >= PAGE_CACHE_SIZE &&
	         ppos <= raend && ppos + filp->f_ralen >= raend) {
/*
 * Add ONE page to max_ahead in order to try to have about the same IO max size
 * as synchronous read-ahead (MAX_READAHEAD + 1)*PAGE_CACHE_SIZE.
 * Compute the position of the last page we have tried to read in order to 
 * begin to read ahead just at the next page.
 */
		raend -= PAGE_CACHE_SIZE;
		if (raend < inode->i_size)
			max_ahead = filp->f_ramax + PAGE_CACHE_SIZE;

		if (max_ahead) {
			filp->f_rawin = filp->f_ralen;
			filp->f_ralen = 0;
			reada_ok      = 2;
		}
	}
/*
 * Try to read ahead pages.
 * We hope that ll_rw_blk() plug/unplug, coalescence, requests sort and the
 * scheduler, will work enough for us to avoid too bad actuals IO requests.
 */
	ahead = 0;
	while (ahead < max_ahead) {
		ahead += PAGE_CACHE_SIZE;
		page_cache = try_to_read_ahead(filp, loff2pgoff(raend + ahead),
						page_cache);
	}
/*
 * If we tried to read ahead some pages,
 * If we tried to read ahead asynchronously,
 *   Try to force unplug of the device in order to start an asynchronous
 *   read IO request.
 * Update the read-ahead context.
 * Store the length of the current read-ahead window.
 * Double the current max read ahead size.
 *   That heuristic avoid to do some large IO for files that are not really
 *   accessed sequentially.
 */
	if (ahead) {
#if 0
		if (reada_ok == 2) {
			run_task_queue(&tq_disk);
		}
#endif

		filp->f_ralen += ahead;
		filp->f_rawin += filp->f_ralen;
		filp->f_raend = raend + ahead + PAGE_CACHE_SIZE;

		filp->f_ramax += filp->f_ramax;

		if (filp->f_ramax > max_readahead)
			filp->f_ramax = max_readahead;

#ifdef PROFILE_READAHEAD
		profile_readahead((reada_ok == 2), filp);
#endif
	}

	return page_cache;
}

/*
 * "descriptor" for what we're up to with a read.
 * This allows us to use the same read code yet
 * have multiple different users of the data that
 * we read from a file.
 *
 * The simplest case just copies the data to user
 * mode.
 */
typedef struct {
	size_t written;
	size_t count;
	char * buf;
	int error;
} read_descriptor_t;

typedef int (*read_actor_t)(read_descriptor_t *, const char *, unsigned long);

/*
 * This is a generic file read routine, and uses the
 * inode->i_op->readpage() function for the actual low-level
 * stuff.
 *
 * This is really ugly. But the goto's actually try to clarify some
 * of the logic when it comes to error handling etc.
 */
static void do_generic_file_read(struct file * filp, loff_t *ppos, read_descriptor_t * desc, read_actor_t actor)
{
	struct dentry *dentry = filp->f_dentry;
	struct inode *inode = dentry->d_inode;
	unsigned long page_cache;
	loff_t pos, posp;
	int reada_ok;
	int max_readahead = get_max_readahead(inode);

	page_cache = 0;

	pos = *ppos;
	posp = pos & PAGE_CACHE_MASK_loff;
/*
 * If the current position is outside the previous read-ahead window, 
 * we reset the current read-ahead context and set read ahead max to zero
 * (will be set to just needed value later),
 * otherwise, we assume that the file accesses are sequential enough to
 * continue read-ahead.
 */
	if (posp > filp->f_raend || posp + filp->f_rawin < filp->f_raend) {
		reada_ok = 0;
		filp->f_raend = 0;
		filp->f_ralen = 0;
		filp->f_ramax = 0;
		filp->f_rawin = 0;
	} else {
		reada_ok = 1;
	}
/*
 * Adjust the current value of read-ahead max.
 * If the read operation stay in the first half page, force no readahead.
 * Otherwise try to increase read ahead max just enough to do the read request.
 * Then, at least MIN_READAHEAD if read ahead is ok,
 * and at most MAX_READAHEAD in all cases.
 */
	if (pos + desc->count <= (loff_t)(PAGE_CACHE_SIZE >> 1)) {
		filp->f_ramax = 0;
	} else {
		loff_t needed;

		needed = ((pos + desc->count) & PAGE_CACHE_MASK) - posp;

		if (filp->f_ramax < needed)
			filp->f_ramax = needed;

		if (reada_ok && filp->f_ramax < MIN_READAHEAD)
				filp->f_ramax = MIN_READAHEAD;
		if (filp->f_ramax > max_readahead)
			filp->f_ramax = max_readahead;
	}

	for (;;) {
		struct page *page, **hash;
		pgoff_t pgoff;

		if (pos >= inode->i_size)
			break;

		/*
		 * Try to find the data in the page cache..
		 */
		pgoff = loff2pgoff(pos);
		hash = page_hash(inode, pgoff);
		page = __find_page(inode, pgoff, *hash);
		if (!page)
			goto no_cached_page;

found_page:
/*
 * Try to read ahead only if the current page is filled or being filled.
 * Otherwise, if we were reading ahead, decrease max read ahead size to
 * the minimum value.
 * In this context, that seems to may happen only on some read error or if 
 * the page has been rewritten.
 */
		if (PageUptodate(page) || PageLocked(page))
			page_cache = generic_file_readahead(reada_ok, filp, inode, pos & PAGE_CACHE_MASK_loff, page, page_cache);
		else if (reada_ok && filp->f_ramax > MIN_READAHEAD)
				filp->f_ramax = MIN_READAHEAD;

		wait_on_page(page);

		if (!PageUptodate(page))
			goto page_read_error;

success:
		/*
		 * Ok, we have the page, it's up-to-date and ok,
		 * so now we can finally copy it to user space...
		 */
	{
		unsigned long offset, nr;

		/* If users can be writing to this page using arbitrary
		 * virtual addresses, take care about potential aliasing
		 * before reading the page on the kernel side.
		 */
		if (inode->i_mmap_shared != NULL)
			flush_dcache_page(page_address(page));

		offset = pos & ~PAGE_CACHE_MASK;
		nr = PAGE_CACHE_SIZE - offset; /* small value */
		if ((loff_t)nr > (inode->i_size - pos))
			nr = inode->i_size - pos;

		/*
		 * The actor routine returns how many bytes were actually used..
		 * NOTE! This may not be the same as how much of a user buffer
		 * we filled up (we may be padding etc), so we can only update
		 * "pos" here (the actor routine has to update the user buffer
		 * pointers and the remaining count).
		 */
		nr = actor(desc, (const char *) (page_address(page) + offset), nr);
		pos += nr;
		page_cache_release(page);
		if (nr && desc->count)
			continue;
		break;
	}

no_cached_page:
		/*
		 * Ok, it wasn't cached, so we need to create a new
		 * page..
		 */
		if (!page_cache) {
			page_cache = page_cache_alloc();
			/*
			 * That could have slept, so go around to the
			 * very beginning..
			 */
			if (page_cache)
				continue;
			desc->error = -ENOMEM;
			break;
		}

		/*
		 * Ok, add the new page to the hash-queues...
		 */
		page = page_cache_entry(page_cache);
		page_cache = 0;
		add_to_page_cache(page, inode, pgoff, hash);

		/*
		 * Error handling is tricky. If we get a read error,
		 * the cached page stays in the cache (but uptodate=0),
		 * and the next process that accesses it will try to
		 * re-read it. This is needed for NFS etc, where the
		 * identity of the reader can decide if we can read the
		 * page or not..
		 */
/*
 * We have to read the page.
 * If we were reading ahead, we had previously tried to read this page,
 * That means that the page has probably been removed from the cache before 
 * the application process needs it, or has been rewritten.
 * Decrease max readahead size to the minimum value in that situation.
 */
		if (reada_ok && filp->f_ramax > MIN_READAHEAD)
			filp->f_ramax = MIN_READAHEAD;

		{
			int error = inode->i_op->readpage(filp, page);
			if (!error)
				goto found_page;
			desc->error = error;
			page_cache_release(page);
			break;
		}

page_read_error:
		/*
		 * We found the page, but it wasn't up-to-date.
		 * Try to re-read it _once_. We do this synchronously,
		 * because this happens only if there were errors.
		 */
		{
			int error = inode->i_op->readpage(filp, page);
			if (!error) {
				wait_on_page(page);
				if (PageUptodate(page) && !PageError(page))
					goto success;
				error = -EIO; /* Some unspecified error occurred.. */
			}
			desc->error = error;
			page_cache_release(page);
			break;
		}
	}

	*ppos = pos;
	filp->f_reada = 1;
	if (page_cache)
		page_cache_free(page_cache);
	UPDATE_ATIME(inode);
}

static int file_read_actor(read_descriptor_t * desc, const char *area, unsigned long size)
{
	unsigned long left;
	unsigned long count = desc->count;

	if (size > count)
		size = count;
	left = __copy_to_user(desc->buf, area, size);
	if (left) {
		size -= left;
		desc->error = -EFAULT;
	}
	desc->count = count - size;
	desc->written += size;
	desc->buf += size;
	return size;
}

#if defined(CONFIG_AIO)

static unsigned long 	kaio_hash_mask = 0;
static kaiocb_t 	**kaio_hash_table;
static int		kaio_hash_table_size;
static int		kaio_number = 1;  	/* id for kaio 		*/
static int		kaio_nr_hashed = 0;

typedef struct {
	spinlock_t kaioq_lock;
	struct semaphore kaiosem;
	struct list_head kaioq;
	int		 kaioqref;
	volatile int	 kaioq_nslaves;
	struct semaphore kaioq_exit_sem;
} kaio_queue_t;

#define KAIO_INFO(task) ((kaio_queue_t *) (task)->kaio_info)

/*
 * AIO HASH TABLE:
 * 	A simple system wide table hashed on kq & aio_id. 
 *	Most of this code adapted from the buffer cache hash table.
 *	All operations protected by kernel lock.
 */
#define _kaio_hashfn(kq, kid)	((unsigned)(((unsigned)kq)^(kid)) & kaio_hash_mask)
#define kaio_hash(kq, kid)	kaio_hash_table[_kaio_hashfn((kq),(kid))]

static void
kaio_init_hash_table(void)
{
	int order = 1;		/* Currently maximum order.. */

	kaio_hash_table_size = (1UL << order) * PAGE_SIZE / sizeof(kaiocb_t *);
	kaio_hash_table = (kaiocb_t  **) __get_free_pages(GFP_ATOMIC, order);
	
	if (!kaio_hash_table)
		panic("Failed to allocate kaio hash table\n");
	memset(kaio_hash_table, 0, kaio_hash_table_size * sizeof(kaiocb_t *));
	kaio_hash_mask = kaio_hash_table_size-1;
}

static void
kaio_add_hash(kaiocb_t *kaiocb)
{
	kaiocb_t **khp;
	kaiocb_t *next;

	kaiocb->kaio_id = kaio_number++; /* XXX: how about overflow? */
	khp = &kaio_hash(current->kaio_info, kaiocb->kaio_id);
	next = *khp;

	kaiocb->kaio_h_next = NULL;
	kaiocb->kaio_h_pprev = NULL;

	if (next) {
		kaiocb->kaio_h_next = next;
		next->kaio_h_pprev = &kaiocb->kaio_h_next;
	}
	*khp = kaiocb;
	kaiocb->kaio_h_pprev = khp;
	kaio_nr_hashed++;
}

static kaiocb_t *
kaio_find_hash(kaio_key_t *kkey)
{		
	kaiocb_t *next;

	next = kaio_hash(KAIO_INFO(current), kkey->kaio_kid);
	for (;;) {
		kaiocb_t *tmp = next;
		if (!next)
			break;
		next = tmp->kaio_h_next;
		if (tmp->kaio_id != kkey->kaio_kid)
			continue;
		assert(tmp->kaio_task->kaio_info == kkey->kaio_kq);
		next = tmp;
		break;
	}
	return next;
}

static inline void
kaio_remove_hash(kaiocb_t *kaiocb)
{
	kaiocb_t **pprev = kaiocb->kaio_h_pprev;
	if (pprev) {
		kaiocb_t *next = kaiocb->kaio_h_next;
		if (next) {
			next->kaio_h_pprev = pprev;
			kaiocb->kaio_h_next = NULL;
		}
		*pprev = next;
		kaiocb->kaio_h_pprev = NULL;
	}
	kaio_nr_hashed--;
}

/*
 * PER-TASK AIO QUEUE OPERATIONS.
 * 	Proceted by a spinlock embedded in the queue.
 */


/*
 * Create: an aio queue for the given task. A q create is attempted as 
 * part of all aio operations, including the slave start operation.
 * The queue is freed as part of exit. 
 */
void
kaio_q_create(struct task_struct *task)
{
	struct task_struct 	*p;
	kaio_queue_t		*kq;

	/* protected by lock_kernel */
	if (KAIO_INFO(task))
		return;

	kq = kmalloc(sizeof(kaio_queue_t), GFP_KERNEL);

	if (!kq) {
		printk("kmalloc KAIO_INFO(task) failed!!\n");
		do_exit(SIGKILL);
	}
	/*
	 * XXX: the following memset is required
	 * in spite of individual member initializaion of
	 * kaio_queue_t. why?
	 */
	memset(&kq->kaioq_lock, 0, sizeof(kaio_queue_t));
	sema_init(&kq->kaiosem, 0);
	sema_init(&kq->kaioq_exit_sem, 0);
	INIT_LIST_HEAD(&kq->kaioq);
	kq->kaioq_nslaves = 0;
	/*
	 * If there are any clones, propage kaio_info.
	 * Following will automatically set current task kaioq & reference.
	 */
	kq->kaioqref = 0;
	for_each_task(p) {
		if (p->mm == current->mm) {
			KAIO_INFO(p) = kq;
			kq->kaioqref++;
		}
	}
	
}


/*
 * Enqueue: now the aio is officially enqueued for I/O, accessible to the slave
 * 	threads, and is also entered into the hash table, accessible to
 * 	user operations such as aio_suspend.
 */
static int
kaio_enqueue(kaio_queue_t *kq, kaiocb_t *kaiocb)
{
	kaiocb->kaio_error = EINPROGRESS;
	if (put_user(kaiocb->kaio_error, &kaiocb->kaio_uaiocb->aio_error))
		return(-EFAULT);
	kaiocb->kaio_wait = 0;
	kaiocb->kaio_suspend_wait = 0;
	kaio_set_time(kaiocb, AIOTIME_REQUEST);
	kaio_add_hash(kaiocb);
	INIT_LIST_HEAD(&kaiocb->kaio_ioq);
	spin_lock(&kq->kaioq_lock);
	list_add(&kaiocb->kaio_ioq, &kq->kaioq);
	spin_unlock(&kq->kaioq_lock);

	up(&kq->kaiosem);
	return 0;
}

/*
 * Wait for an aio and dequeue. Normally, kaiosem counts the number
 * of entries available in the queue; however, kaio_cancel can
 * dequeue entries as well, and it does not down the semaphore,
 * because there is no easy way to synchronize this action with kaio_dequeue.
 * Instead, the dequeue operation simply retries if it finds the queue
 * empty eventhough the semaphore count was non-zero (i.e. the down succeded).
 * The down() is interrupted when the parent sends a signal when it exits.
 */
static kaiocb_t *
kaio_dequeue(kaio_queue_t *kq)
{
	struct list_head *tmp;

retry:
	if (down_interruptible(&kq->kaiosem))
		return 0;

	spin_lock(&kq->kaioq_lock);
	tmp = kq->kaioq.prev;
	if (tmp != &kq->kaioq) {
		list_del(tmp);
		spin_unlock(&kq->kaioq_lock);
		INIT_LIST_HEAD(tmp);
		return(list_entry(tmp, struct kaiocb, kaio_ioq));
	} else {
		spin_unlock(&kq->kaioq_lock);
		goto retry;
	}
}

static int
kaio_dequeue_kaiocb(kaio_queue_t *kq, kaiocb_t *kaiocb)
{
	spin_lock(&kq->kaioq_lock);
	if (kaiocb->kaio_ioq.prev == &kaiocb->kaio_ioq) {
		assert(kaiocb->kaio_ioq.next == &kaiocb->kaio_ioq);
		spin_unlock(&kq->kaioq_lock);
		return 0;
	}
	list_del(&kaiocb->kaio_ioq);
	INIT_LIST_HEAD(&kaiocb->kaio_ioq);
	spin_unlock(&kq->kaioq_lock);
	return 1;
}

static kaiocb_t *
kaio_dequeue_fd(kaio_queue_t *kq, struct file *filp)
{
	struct list_head *tmp;
	kaiocb_t *kaiocb;

	spin_lock(&kq->kaioq_lock);
	tmp = kq->kaioq.next; /* start from first of list; why not? */
	while (tmp != &kq->kaioq) {
		kaiocb = list_entry(tmp, struct kaiocb, kaio_ioq);
		if (kaiocb->kaio_filp == filp) {
			list_del(tmp);
			spin_unlock(&kq->kaioq_lock);
			INIT_LIST_HEAD(tmp);
			return(kaiocb);
		}
		tmp = tmp->next;
	}
	spin_unlock(&kq->kaioq_lock);
	return 0;
}

/*
 * Boot time initialization.
 */
void
kaio_init(void)
{
	kaio_init_hash_table();
}

static inline unsigned long
kaio_time(void)
{
	struct timeval tv;

	do_gettimeofday(&tv);
	return (tv.tv_sec * 1000000 + tv.tv_usec);
}

static void
send_signal(sigevent_t *sigev, struct task_struct *task)
{
	if (sigev->sigev_notify == SIGEV_SIGNAL) {
		struct siginfo sinfo;
		
		memset(&sinfo, 0, sizeof(sinfo));
		sinfo.si_signo = sigev->sigev_signo;
		sinfo.si_code = SI_ASYNCIO;
		sinfo.si_value = sigev->sigev_value;

		send_sig_info(sigev->sigev_signo, &sinfo, task);
	}
}

static int
lio_complete(liocb_t *liocb, int n)
{
	spin_lock(&liocb->lio_lock);
	liocb->lio_done += n;
	if (liocb->lio_nent == liocb->lio_done) {
		wake_up(&liocb->lio_wait);
		send_signal(&liocb->lio_sigevent, liocb->lio_task);
		if (liocb->lio_mode == LIO_NOWAIT)
			kfree(liocb);
		return 1;
	}
	spin_unlock(&liocb->lio_lock);
	return 0;
}

static int
kaio_io_complete(kaiocb_t *kaiocb, int nbytes, int ret, int gc)
{
	aiocb_t	*uaiocb = kaiocb->kaio_uaiocb;
	liocb_t	*liocb = kaiocb->kaio_liocb;
	int	i;

	lock_kernel();

	fput(kaiocb->kaio_filp);
	if (kaiocb->kaio_pages) {

		for (i = 0; i < kaiocb->kaio_npages; i++)
			__free_page(kaiocb->kaio_pages[i]);
		kfree(kaiocb->kaio_pages);
	}

#ifdef KAIO_STATS
	if (copy_to_user(&uaiocb->aio_times, &kaiocb->kaio_times,
			sizeof(unsigned long) * AIO_TIMES))
	{
		ret = -EFAULT;
	} else
#endif
	if (put_user(nbytes, &uaiocb->aio_return))
		ret = -EFAULT;
	/* aio_error does not go through syscall so -ret */
	else if (put_user(-ret, &uaiocb->aio_error))
		ret = -EFAULT;
	else {
		kaiocb->kaio_error = -ret;
		send_signal(&kaiocb->kaio_sigevent, kaiocb->kaio_task);
		if (kaiocb->kaio_suspend_wait)
			wake_up(kaiocb->kaio_suspend_wait);
		else
			gc = 1;
		if (gc || (liocb && liocb->lio_mode == LIO_WAIT)) {
			kaio_remove_hash(kaiocb);
			kfree(kaiocb);
		}
		if (liocb)
			lio_complete(liocb, 1);
		ret = 0;
	}
	unlock_kernel();

	return ret;
}

static int
kaio_slave_suspend(kaiocb_t *kaiocb)
{
	off_t pos;
	int i, ret = 0, nbytes, isize, npending;
	struct page *page;

	nbytes = kaiocb->kaio_nbytes;
	isize = kaiocb->kaio_filp->f_dentry->d_inode->i_size;
	atomic_set(&kaiocb->kaio_pending, 0);
	for (i = 0, npending = 0; i < kaiocb->kaio_npages; i++) {
		page = kaiocb->kaio_pages[i];
		wait_on_page(page);
		/*
		 * Ok, we have the page. 
		 */
		if (!PageUptodate(page) || PageError(page)) {
			ret = -EIO; /* Some unspecfied error */
			goto error;
		}
	}
	kaio_set_time(kaiocb, AIOTIME_IO);

	pos = kaiocb->kaio_offset;
	i = 0;
	page = kaiocb->kaio_pages[0];
	assert(nbytes);

	while  (nbytes && pos < isize) {
		unsigned long offset, nr, left;

		offset = pos & ~PAGE_MASK;
		nr = PAGE_SIZE - offset;
		if (nr > isize - pos)
			nr = isize - pos;
		if (nbytes < nr)
			nr = nbytes;

		left = __copy_to_user(kaiocb->kaio_buf,
				(char *) page_address(page) + offset, nr);
		nbytes -= (nr - left);
		if (left) {
			ret = -EFAULT;
			goto error; 
		}
		pos += nr;
		kaiocb->kaio_buf = (char *)kaiocb->kaio_buf + nr;
		if ((pos & ~PAGE_MASK) == 0) { /* beginning of next page */
			i++;
			page = kaiocb->kaio_pages[i];
		}
	}
	assert(!nbytes || pos == isize);
	kaio_set_time(kaiocb, AIOTIME_COPY);

error:

	/* nbytes contains remainder unread, for eg., due to isize limit */
	ret = kaio_io_complete(kaiocb, kaiocb->kaio_nbytes - nbytes, ret, 0);

	return ret;
}

#if defined(CONFIG_RAW)
static int
do_raw_wait(kaiocb_t *kaiocb)
{
	int diag;

	do {
		diag = kaio_raw_wait(kaiocb->kaio_scp, &kaiocb->kaio_iobuf);
	} while (diag == -EAGAIN);
	diag = kaio_io_complete(kaiocb, diag, (diag > 0 ? 0 : diag), 0);
	return diag;
}

static int
do_kaio_raw_rw(int cmd, struct file *filp, struct kaiocb *kaiocb)
{
	int diag;
	void *scp;

	unlock_kernel();		/* XXX - await sfc changes */
	kaiocb->kaio_sem = MUTEX_LOCKED;
	scp = kaio_raw_rw((cmd == AIOCMD_READ ? READ : WRITE), filp,
			kaiocb->kaio_buf, kaiocb->kaio_nbytes,
			&kaiocb->kaio_offset, &diag,
			&kaiocb->kaio_sem, &kaiocb->kaio_iobuf);
	kaiocb->kaio_scp = scp;
	lock_kernel();			/* XXX - await sfc changes */
	if (diag == 0)
		diag = kaio_enqueue(KAIO_INFO(current), kaiocb);
	return diag;
}
#endif /* CONFIG_RAW */


int
kaio_suspend(aiocb_t *list[], int nent, struct timespec *utimeout, int nwait)
{
	int 			ret = 0, i, valid, pending;
	struct timespec 	timeout;
	aiocb_t			*uaiocb = 0;
	kaio_key_t		kkey;
	kaiocb_t 		*kaiocb, **kaiocbs, *kaiocb_array[256];
	struct wait_queue	wait, *wq = NULL;

	if (utimeout && copy_from_user(&timeout, (struct timespec *)utimeout,
				    sizeof(struct timespec)))
	{
		return -EFAULT;
	}
	/*
	 * For more normal requests, allocate on stack.
	 */
	if (nent > 256) {
		kaiocbs = kmalloc(sizeof(kaiocb_t *) * nent, GFP_KERNEL);
		if (!kaiocbs)
			return -EAGAIN;
	} else
		kaiocbs = kaiocb_array;

	wait.task = current;
	wait.next = NULL;
	add_wait_queue(&wq, &wait);

	/*
	 * Go through the list, if nwait done, return; otherwise
	 * mark kaiocb so it will wake this thread when done.
	 */
	for (i = 0, valid = 0; nwait > 0 && i < nent; i++) {
		if (get_user(uaiocb, &list[i])) {
			ret = -EFAULT;
			goto suspend_error;
		}
		if (!uaiocb)
			continue;
		/*
		 * Check to see if done, otherwise put this
		 * thread on the list of things to wake up.
		 */
		if (copy_from_user(&kkey, &uaiocb->aio_key,
			sizeof(kaio_key_t)))
		{
			ret = -EFAULT;
			goto suspend_error;
		}
		if ((kaiocb = kaio_find_hash(&kkey)) == 0) {
			if (get_user(ret, &uaiocb->aio_error)) {
				ret = -EFAULT;
				goto suspend_error;
			}
			/*
			 * Assume that user gave a valid aiocb,
			 * and that it is done.
			 */
			nwait--;
		} else {
			if (kaiocb->kaio_suspend_wait) {
				/*
				 * Don't allow two different processes to
				 * wait on the same I/O request.
				 */
				ret = -EAGAIN;
				goto suspend_error;
			}
			kaiocb->kaio_suspend_wait = &wq;
			if (kaiocb->kaio_error != EINPROGRESS) {
				printk("Found kaiocb 0x%p in hash but not"
					"in progress (kaio_error %d)\n",
					kaiocb, kaiocb->kaio_error);
				panic("Kaio hash inconsistent");
			}
			kaiocbs[valid++] = kaiocb;
		}
	}
	/*
	 * Pending keeps track of how many I/Os  are being
	 * wait upon. Valid is the index into kaiocbs where
	 * the kaiocb of the last pending I/O is stored.
	 * After sleeping & waking up on any one of the pending
	 * I/O to complete, we scan the kaiocbs, garbage collect
	 * all complete kaiocb & put a null entry in kaiocbs[].
	 */
	pending = valid;
	while (nwait > 0 && pending) {
		if (signal_pending(current)) {
			ret = -ERESTARTSYS;
			goto suspend_error; 
		}
		current->state = TASK_INTERRUPTIBLE;
		if (utimeout)
			(void) schedule_timeout(timespec_to_jiffies(&timeout));
		else
			schedule();
		current->state = TASK_RUNNING;
		if (signal_pending(current)) {
			ret = -ERESTARTSYS;
			goto suspend_error; 
		}
		/*
		 * See how many more done & free kaiocb's as necessary.
		 */
		for (i = 0; i < valid; i++) {
			if (kaiocbs[i] &&
				kaiocbs[i]->kaio_error != EINPROGRESS)
			{
				kaio_remove_hash(kaiocbs[i]);
				kfree(kaiocbs[i]);
				nwait--;
				pending--;
				kaiocbs[i] = 0;
			}
		}
	}

suspend_error:
	/*
	 * Unmark & free kaiocb's as necessary.
	 */
	for (i = 0; i < valid; i++) {
		if (!kaiocbs[i])
			continue;
		kaiocbs[i]->kaio_suspend_wait = 0;
		if (kaiocbs[i]->kaio_error != EINPROGRESS) {
			kaio_remove_hash(kaiocbs[i]);
			kfree(kaiocbs[i]);
		}
	}
	if (kaiocbs != kaiocb_array)
		kfree(kaiocbs);
	return(ret);
}

/*
 * kaio_cancel: if given kkey, cancel a specific aio; otherwise
 * 		cancel all aios for given fd. All successful canceled
 *		aios will have -1 return & ECANCELED as aio_error.
 *		kaio_io_complete handles normal signal delivery and
 *		any listio that the aio may be part of.
 *	kernel lock held on call.
 */
int
kaio_cancel(int fd, kaio_key_t *kkey)
{
	kaiocb_t 	*kaiocb;
	struct file 	*filp;
	int		ret = 0;

	if (!(filp = fget(fd)))
		return -EBADF;

	if (kkey) {
		if ((kaiocb = kaio_find_hash(kkey)) == 0) {
			ret = -EINVAL;
			goto cancel_out;
		}

		if (filp != kaiocb->kaio_filp) {
			ret = -EBADF;
			goto cancel_out;
		}

		if (!kaio_dequeue_kaiocb(KAIO_INFO(current), kaiocb)) {
			ret = AIO_NOTCANCELED;
			goto cancel_out;
		}
		ret = kaio_io_complete(kaiocb, -1, -ECANCELED, 1);
		if (ret == 0)
			ret = AIO_CANCELED;
	} else {
		/* cancel all associated with fd */
		if ((kaiocb = kaio_dequeue_fd(KAIO_INFO(current), filp)) == 0) {
			ret = AIO_ALLDONE;
			goto cancel_out;
		}
		do {
			/* XXX: what if error occurs */
			kaio_io_complete(kaiocb, -1, -ECANCELED, 1);
		} while ((kaiocb = kaio_dequeue_fd(KAIO_INFO(current), filp)) != 0);
	}
cancel_out:
	fput(filp);
	return ret;
}

void
kaio_fork(struct task_struct *p, int clone_flags)
{
	/* Assumes kernel lock held */

	/*
	 * If the new thread is not sharing VM with parent,
	 * then it can't share AIO state.
	 */
	if (clone_flags & CLONE_VM) {
		KAIO_INFO(p) = KAIO_INFO(current);
		KAIO_INFO(p)->kaioqref++;
	} else
		p->kaio_info = 0;
}

/*
 * kaio_exit:
 *	Called by generic exit when a thread has async I/O (kaio_info).
 *	This routine does the following:
 * 	(1) kills the slave threads, by signalling and
 * 	     waiting for them to be reaped.
 *	(2) Removes any queued aios (these may not have been intiated at all).
 *	(3) Removes any remaining aios for this thread from the hash (these
 *		have completed but not reaped through aio_suspend).
 *	All actions protected by kernel lock.
 */
void
kaio_exit(void)
{
	struct task_struct 	*p;
	kaiocb_t 		**khpp;
	kaio_queue_t		*kq;
	struct list_head 	*tmp;
	kaiocb_t 		*curr, *next;

	lock_kernel();

	/* kill all slaves if last clone */

	kq = KAIO_INFO(current);
	if (kq->kaioqref != 1) {
		kq->kaioqref--;
		unlock_kernel();
		return;
	}
	/*
	 * Last reference, kill all slaves & cleanup pending I/Os.
	 */

	for_each_task(p) {
		/*
		 * Look for slaves by name.
		 */
		if (p->mm == current->mm &&
		     strncmp(&p->comm[0], KAIO_SLAVE_NAME,
			     KAIO_SLAVE_NAMELEN) == 0)
		{ 
			send_sig(SIGKILL, p, 1);
		}
	}
	/* kernel lock held for the test below */
	while (kq->kaioq_nslaves != 0)
		down(&kq->kaioq_exit_sem);
	/*
	 * Any queued aios can be simply "completed" & garbage collected. 
	 */
	for (tmp = kq->kaioq.prev; tmp != &kq->kaioq; tmp = kq->kaioq.prev) {
		list_del(tmp);
		kaio_io_complete(list_entry(tmp, struct kaiocb, kaio_ioq),
					-1, -1, 1);
	}
	/*
	 * Free up all hashed aios for this pid; these are already complete,
	 * but pending aio_suspend.
	 */
	for (khpp = &kaio_hash_table[0];
		khpp < &kaio_hash_table[kaio_hash_table_size]; khpp++)
	{

		for  (curr = *khpp; curr; curr = next) {
			next = curr->kaio_h_next;
			if (curr->kaio_kq == kq) {
				kaio_remove_hash(curr);
				kfree(curr);
			}
		}
	}
	kfree(kq);
	unlock_kernel();
}

#define KAIO_DO_FOP(op, k)  {						      \
	int ret;							      \
	lock_kernel(); 							      \
	(k)->kaio_filp->f_pos = (k)->kaio_offset;                             \
	ret = (k)->kaio_filp->f_op-> ## op ((k)->kaio_filp,                   \
		(k)->kaio_buf, (k)->kaio_nbytes, &(k)->kaio_filp->f_pos);     \
	kaio_io_complete((k), ret < 0 ? 0 : ret, ret > 0 ? 0 : ret, 0);	      \
	unlock_kernel(); 						      \
}


static void
kaio_set_slave_name(void)
{
	char newcomm[16];

	sprintf(newcomm, "%s%05d", KAIO_SLAVE_NAME, current->p_pptr->pid);
	strcpy(current->comm, newcomm);
}

/*
 * kaio_slave:
 * 	Entry point for slave threads that service queued I/O requests.
 * For each task that employs async I/O there may be several slave threads.
 * 
 *	The main function of the slave threads is to wait & dequeue an I/O
 * request and serice it. Requests may be split-phase with a seperate initiate
 * and wait routine, or with a unified routine that does both.  Not all
 * file systems support split-phase requests; only raw I/O directly support it.
 * Additionally, for file systems that employ generic_read we can use
 * the recoded kaio_generic_read & associated wait (kaio_slave_suspend).
 *	
 *	Slave threads are explicitly started by the library, but are killed
 * when the "parent" task (that initiates I/O) exits (kaio_exit).
 */
void
kaio_slave()
{
	kaiocb_t *kaiocb;
	kaio_queue_t *kq;

	if (current->p_pptr->pid == 1) { /* too late */
		unlock_kernel();
		return;
	}

	kaio_set_slave_name();
	kq = KAIO_INFO(current->p_pptr); /* service parent's kaio queue */
	KAIO_INFO(current) = 0; /* so kaio_exit does not trip */
	kq->kaioqref--;
	kq->kaioq_nslaves++;
	unlock_kernel();

	while ((kaiocb = kaio_dequeue(kq))) { /* blocks */
#if defined(CONFIG_RAW)
		if (kaio_is_raw(kaiocb->kaio_filp))
			do_raw_wait(kaiocb);
		else
#endif
		{
			switch (kaiocb->kaio_cmd) {
			case AIOCMD_GENERIC_READ:
				kaio_slave_suspend(kaiocb);
				break;
			case AIOCMD_READ:
				KAIO_DO_FOP(read, kaiocb);
				break;
			case AIOCMD_WRITE: {
				struct semaphore *isem;

				/*
				 * Have to keep a local reference to the
				 * semaphore, kaiocb cannot be touched after
				 * the FOP which does an I/O complete,
				 * which in turn can free kaiocb.
				 */
				isem =
				  &kaiocb->kaio_filp->f_dentry->d_inode->i_sem;
				down(isem);
				KAIO_DO_FOP(write, kaiocb);
				up(isem);

				break;
			 }
			 default:
				kaio_io_complete(kaiocb, 0, -EINVAL, 0);
			}
		}
	}
	lock_kernel();
	kq->kaioq_nslaves--;
	up(&kq->kaioq_exit_sem);
	unlock_kernel();
}

/*
 * This is an asynchronous generic file read routine, and uses the
 * inode->i_op->readpage() function for the actual low-level stuff.
 */
static int
do_kaio_generic_read(struct file * filp, kaiocb_t *kaiocb)
{
	struct dentry *dentry = filp->f_dentry;
	struct inode *inode = dentry->d_inode;
	size_t pos, page_cache;
	struct page **hash, *page;
	int i, ret, npages;

	pos = kaiocb->kaio_offset;
	if (pos >= inode->i_size) {
		/*
		 * No I/O possible; specified offset beyond file end
		 */
		return -EINVAL;
	}

	i = pos + kaiocb->kaio_nbytes - 1; /* last byte in I/O */
	if (i > inode->i_size)
		i = inode->i_size;
	npages = pageof(i) - pageof(pos) + 1;
	/*
	 * Allocate space to hold page pointers.
	 */
	kaiocb->kaio_pages = kmalloc(sizeof(struct page *) * npages, GFP_KERNEL);
	if (!kaiocb->kaio_pages) 
		return -EAGAIN;
	page_cache = 0;
	pos = pageof(pos) * PAGE_SIZE;
	kaiocb->kaio_pchits = 0;
	ret = 0;
	for (i = 0; i < npages; i++, pos += PAGE_SIZE) {

retry:
		/*
		 * Try to find the data in the page cache..
		 */
		hash = page_hash(inode, pos & PAGE_MASK);
		page = __find_page(inode, pos & PAGE_MASK, *hash);
		if (page) {
			kaiocb->kaio_pchits++;
			kaiocb->kaio_pages[i] = page;
			continue;
		}

		/*
		 * Ok, it wasn't cached, so we need to create a new
		 * page.. or reuse a page previously allocated,
		 * and did not use in retry path below (found page in
		 * hash after allocation).
		 */
		if (!page_cache) {
			page_cache = __get_free_page(GFP_USER);
			/*
			 * That could have slept, so go around to the
			 * very beginning..
			 */
			if (page_cache)
				goto retry;
			ret = -ENOMEM;
			break;
		}

		/*
		 * Ok, add the new page to the page cache.
		 */
		page = mem_map + MAP_NR(page_cache);
		page_cache = 0;
		add_to_page_cache(page, inode, pos & PAGE_MASK, hash);
		kaiocb->kaio_pages[i] = page;

		/*
		 * Fire off the I/O; read page may also add page's
		 * bufferheads to buffer cache.
		 */
		if ((ret = inode->i_op->readpage(filp, page)))
			break;
		run_task_queue(&tq_disk); /* XXX - improves perf. but why ??*/
	}
	if (!ret) {
		kaiocb->kaio_npages = npages;
		ret = kaio_enqueue(KAIO_INFO(current), kaiocb);
	}
	if (ret) { /* recheck return from kaio_enqueue */
		for (i--; i >=0; i--) {
			assert(kaiocb->kaio_pages[i]);
			__free_page(kaiocb->kaio_pages[i]);
		}
		kfree(kaiocb->kaio_pages);
	}
	if (page_cache)
		free_page(page_cache);
	UPDATE_ATIME(inode);
	return ret;
}

/*
 * Initiate buffered async I/O (e.g. file systems).
 *
 * Kaiocb describing the I/O is queued as part of a per-task queue (KAIO_INFO).
 * I/O is actually initiated only in the case of generic_read. For all
 * other cases, slave threads initiate and wait for queued requests.
 *
 */
static int
do_kaio_fs_rw(int cmd, struct file *filp, struct kaiocb *kaiocb)
{
	/*
	 * if read request & the fop operation is generic,
	 * use our own parallel async version; for all other
	 * reads & for writes, asynchronize by queuing f_op to be
	 * executed directly by slave.
	 */
	if (cmd == AIOCMD_READ && filp->f_op->read == generic_file_read) {
		kaiocb->kaio_cmd = AIOCMD_GENERIC_READ;
		return(do_kaio_generic_read(filp, kaiocb));
	} else {
		return(kaio_enqueue(KAIO_INFO(current), kaiocb));
	}
}

/*
 * kaio_rw: Initiate async I/O read or write.
 *
 *	cmd 	= AIOCMD_{READ,WRITE}; ignored for listio.
 *	uaiocb 	= user space address of aiocb describing request.
 *	liocb 	= if not null then part of a list io. 
 *
 * Kernel lock held across this function. If I/O successfully queued,
 * kaiocb holds reference to file (with fget, correspoding fput will be done
 * at io completion/cancellation/exit). Kaiocb also holds references to the
 * requesting task, the uaiocb (for transmitting back return/error values),
 * and liocb, if any. A unique id, representing the  kaiocb in a hash table,
 * is "returned" back to user on successful queueing.
 *
 */
int
kaio_rw(int cmd, aiocb_t *uaiocb, liocb_t *liocb)
{
	struct file 	*filp = NULL;
	aiocb_t		aiocb;
	int 		ret;
	kaiocb_t 	*kaiocb = NULL;


	if (copy_from_user(&aiocb, uaiocb, sizeof(aiocb_t)))
		return -EFAULT;

	if (kaio_nr_hashed > CONFIG_AIO_MAX) {
		ret = -EAGAIN;
		goto rw_error;
	}
	if (liocb) {
		switch (aiocb.aio_lio_opcode) {
		case LIO_READ:
			cmd = AIOCMD_READ;
			break;
		case LIO_WRITE:
			cmd = AIOCMD_WRITE;
			break;
		case LIO_NOP:
			/*
			 * No easy way to return to caller an
			 * indication to ignore; so, alternately,
			 * mark as complete.
			 */
			return(lio_complete(liocb, 1));
		default:
			ret =  -EINVAL;
			goto rw_error;
		}
	}
	if (!(filp = fget(aiocb.aio_fildes))) {
		ret = -EBADF;
		goto rw_error;
	}
	if (!(filp->f_mode &
		(cmd == AIOCMD_READ ? FMODE_READ: FMODE_WRITE)))
	{
		ret = -EBADF;
		goto rw_error;
	}
	ret = locks_verify_area(cmd == AIOCMD_READ ?
				FLOCK_VERIFY_READ : FLOCK_VERIFY_WRITE,
				filp->f_dentry->d_inode, filp,
				aiocb.aio_offset, aiocb.aio_nbytes);
	if (ret)
		goto rw_error;

	if (cmd == AIOCMD_READ &&
		!access_ok(VERIFY_WRITE, aiocb.aio_buf, aiocb.aio_nbytes))
	{
		ret = -EFAULT;
		goto rw_error;
	}
	if (aiocb.aio_nbytes) {

		kaiocb = kmalloc(sizeof(kaiocb_t), GFP_KERNEL);
		if (!kaiocb) {
			ret = -EAGAIN;
			goto rw_error;
		}
		kaiocb->kaio_nbytes = aiocb.aio_nbytes;
		kaiocb->kaio_fildes = aiocb.aio_fildes;
		kaiocb->kaio_buf = aiocb.aio_buf;
		kaiocb->kaio_offset = aiocb.aio_offset;
		kaiocb->kaio_sigevent = aiocb.aio_sigevent;
		kaiocb->kaio_task = current;
		kaiocb->kaio_kq = KAIO_INFO(current);
		kaiocb->kaio_filp = filp;
		kaiocb->kaio_uaiocb = uaiocb;
		kaiocb->kaio_cmd = cmd;
		kaiocb->kaio_liocb = liocb;
		kaiocb->kaio_pages = 0;
#if defined(CONFIG_RAW)
		if (kaio_is_raw(filp))
			ret = do_kaio_raw_rw(cmd, filp, kaiocb);
		else
#endif
			ret = do_kaio_fs_rw(cmd, filp, kaiocb);
	}
	if (!ret) {
		kaio_key_t kkey;

		kkey.kaio_kid = kaiocb->kaio_id;
		kkey.kaio_kq = current->kaio_info;

		if (copy_to_user(&uaiocb->aio_key, &kkey, sizeof(kkey)))
			ret = -EFAULT;
	}
rw_error:
	if (ret) {
		if (filp)
			fput(filp);
		if (kaiocb)
			kfree(kaiocb);
	}
	if (liocb && ret) {
		/*
		 * Have to set individual errors for list io
		 */
		 (void) put_user(-ret, &uaiocb->aio_error);
	}
	return ret;
}


static void
kaio_set_list_errors(aiocb_t *list[], int nent, int error)
{
	int	i;
	aiocb_t *uaiocbp;

	/*
	 * Have to set individual errors for each aiocb.
	 */
	for (i = 0; i < nent; i++)
		if (get_user(uaiocbp, &list[i]) == 0 && uaiocbp)
			 (void) put_user(-error, &uaiocbp->aio_error);
}

int
kaio_list_io(int mode, struct aiocb *list[], int nent, sigevent_t *sig)
{
	liocb_t *liocb;
	aiocb_t *aiocbp;
	int	i = 0, ret = 0, ignore = 0, error = 0;

	liocb = kmalloc(sizeof(liocb_t), GFP_KERNEL);
	if (!liocb) {
		ret = -EAGAIN;
		kaio_set_list_errors(list, nent, ret);
		return ret;
	}

	liocb->lio_lock = SPIN_LOCK_UNLOCKED;
	liocb->lio_nent = nent;
	liocb->lio_done = 0;
	liocb->lio_mode = mode;
	init_waitqueue(&liocb->lio_wait);
	liocb->lio_task = current;
	if (sig) {
		if (copy_from_user(&liocb->lio_sigevent,
			sig, sizeof(sigevent_t)))
		{
			ret =  -EFAULT;
			kaio_set_list_errors(list, nent, ret);
			error = nent; /* all are in error */
			goto lio_error;
		}
	} else
		liocb->lio_sigevent.sigev_notify = SIGEV_NONE;
	for ( ; i < nent; i++) {
		if (get_user(aiocbp, &list[i])) {
			ret = -EFAULT;
			error++;
			continue;
		}
		if (!aiocbp) {
			ignore++;
			continue;
		}
		/* 0 == cmd for kaio_rw derived from aiocbp */
		if ((ret = kaio_rw(0, aiocbp, liocb)))
			error++;
	}
	if (!error) {
		if (!lio_complete(liocb, ignore) && mode == LIO_WAIT)
			interruptible_sleep_on(&liocb->lio_wait);
		if (mode == LIO_WAIT)
			kfree(liocb);
		return 0;
	}
lio_error:
	if (!lio_complete(liocb, (ignore + error))) {
		/* pending I/O's; but we are not interested */
		liocb->lio_mode = LIO_NOWAIT;
		liocb->lio_sigevent.sigev_notify = SIGEV_NONE;
	} else if (mode != LIO_NOWAIT) {
		/* if no wait then lio_complete frees liocb */
		kfree(liocb);
	}
	return ret;
}
#endif /* CONFIG_AIO */

/*
 * This is the "read()" routine for all filesystems
 * that can use the page cache directly.
 */
ssize_t generic_file_read(struct file * filp, char * buf, size_t count, loff_t *ppos)
{
	ssize_t retval;
	struct inode *inode = filp->f_dentry->d_inode;

	if (((ssize_t) count) < 0)
		return -EINVAL;

	retval = -EFAULT;
	if (access_ok(VERIFY_WRITE, buf, count)) {
		retval = 0;

		/* L-F-S spec 2.2.1.25: */
		if (!(filp->f_flags & O_LARGEFILE) &&
		    S_ISREG(inode->i_mode) &&
		    (*ppos < inode->i_size) && count) {
			if (*ppos >= 0x7fffffff) /* pos@2G forbidden */
				return -EOVERFLOW;
			if (*ppos + count > 0x7fffffff)
				/* Read only until end of allowed region */
				count = 0x7fffffff - *ppos;
		}

		if (count) {
			read_descriptor_t desc;

			desc.written = 0;
			desc.count = count;
			desc.buf = buf;
			desc.error = 0;
			do_generic_file_read(filp, ppos, &desc, file_read_actor);

			retval = desc.written;
			if (!retval)
				retval = desc.error;
		}
	}
	return retval;
}

static int file_send_actor(read_descriptor_t * desc, const char *area, unsigned long size)
{
	ssize_t written;
	unsigned long count = desc->count;
	struct file *file = (struct file *) desc->buf;
	struct inode *inode = file->f_dentry->d_inode;
	mm_segment_t old_fs;

	if (size > count)
		size = count;
	fs_down(&inode->i_sem);
	old_fs = get_fs();
	set_fs(KERNEL_DS);
	written = file->f_op->write(file, area, size, &file->f_pos);
	set_fs(old_fs);
	fs_up(&inode->i_sem);
	if (written < 0) {
		desc->error = written;
		written = 0;
	}
	desc->count = count - written;
	desc->written += written;
	return written;
}

asmlinkage ssize_t sys_sendfile(int out_fd, int in_fd, off_t *offset, size_t count)
{
	ssize_t retval;
	struct file * in_file, * out_file;
	struct inode * in_inode, * out_inode;

	if (((ssize_t) count) < 0)
		return -EINVAL;

	lock_kernel();

	/*
	 * Get input file, and verify that it is ok..
	 */
	retval = -EBADF;
	in_file = fget(in_fd);
	if (!in_file)
		goto out;
	if (!(in_file->f_mode & FMODE_READ))
		goto fput_in;
	retval = -EINVAL;
	in_inode = in_file->f_dentry->d_inode;
	if (!in_inode)
		goto fput_in;
	if (!in_inode->i_op || !in_inode->i_op->readpage)
		goto fput_in;
	retval = locks_verify_area(FLOCK_VERIFY_READ, in_inode, in_file, in_file->f_pos, count);
	if (retval)
		goto fput_in;

	/*
	 * Get output file, and verify that it is ok..
	 */
	retval = -EBADF;
	out_file = fget(out_fd);
	if (!out_file)
		goto fput_in;
	if (!(out_file->f_mode & FMODE_WRITE))
		goto fput_out;
	retval = -EINVAL;
	if (!out_file->f_op || !out_file->f_op->write)
		goto fput_out;
	out_inode = out_file->f_dentry->d_inode;
	if (!out_inode)
		goto fput_out;
	retval = locks_verify_area(FLOCK_VERIFY_WRITE, out_inode, out_file, out_file->f_pos, count);
	if (retval)
		goto fput_out;

	retval = 0;
	if (count) {
		read_descriptor_t desc;
		loff_t pos = 0, *ppos;

		retval = -EFAULT;
		ppos = &in_file->f_pos;
		if (offset) {
			if (get_user(pos, offset))
				goto fput_out;
			ppos = &pos;
		}

		desc.written = 0;
		desc.count = count;
		desc.buf = (char *) out_file;
		desc.error = 0;
		do_generic_file_read(in_file, ppos, &desc, file_send_actor);

		retval = desc.written;
		if (!retval)
			retval = desc.error;
		if (offset)
			put_user(pos, offset);
	}


fput_out:
	fput(out_file);
fput_in:
	fput(in_file);
out:
	unlock_kernel();
	return retval;
}

/*
 * Semantics for shared and private memory areas are different past the end
 * of the file. A shared mapping past the last page of the file is an error
 * and results in a SIGBUS, while a private mapping just maps in a zero page.
 *
 * The goto's are kind of ugly, but this streamlines the normal case of having
 * it in the page cache, and handles the special cases reasonably without
 * having a lot of duplicated code.
 *
 * WSH 06/04/97: fixed a memory leak and moved the allocation of new_page
 * ahead of the wait if we're sure to need it.
 */
static unsigned long filemap_nopage(struct vm_area_struct * area, unsigned long address, int no_share)
{
	struct file * file = area->vm_file;
	struct dentry * dentry = file->f_dentry;
	struct inode * inode = dentry->d_inode;
	loff_t offset;
	pgoff_t pgoff, reada;
	int i;
	struct page * page, **hash;
	unsigned long old_page, new_page;

	new_page = 0;
	offset = ((loff_t)((address & PAGE_MASK) - area->vm_start) +
		  area->vm_offset);

	if (offset >= inode->i_size && (area->vm_flags & VM_SHARED) && area->vm_mm == current->mm)
		goto no_page;

	/*
	 * Do we have something in the page cache already?
	 */
	pgoff = loff2pgoff(offset);
	hash = page_hash(inode, pgoff);
	page = __find_page(inode, pgoff, *hash);
	if (!page)
		goto no_cached_page;

found_page:
	/*
	 * Ok, found a page in the page cache, now we need to check
	 * that it's up-to-date.  First check whether we'll need an
	 * extra page -- better to overlap the allocation with the I/O.
	 */
	if (no_share && !new_page) {
		new_page = page_cache_alloc();
		if (!new_page)
			goto release_and_oom;
	}

	if (PageLocked(page))
		goto page_locked_wait;
	if (!PageUptodate(page))
		goto page_read_error;

success:
	/*
	 * Found the page, need to check sharing and possibly
	 * copy it over to another page..
	 */
	old_page = page_address(page);
	if (!no_share) {
		/*
		 * Ok, we can share the cached page directly.. Get rid
		 * of any potential extra pages.
		 */
		if (new_page)
			page_cache_free(new_page);

		flush_page_to_ram(old_page);
		return old_page;
	}

	/*
	 * No sharing ... copy to the new page.
	 */
	copy_page(new_page, old_page);
	flush_page_to_ram(new_page);
	page_cache_release(page);
	return new_page;

no_cached_page:
	/*
	 * Try to read in an entire cluster at once.
	 */
	reada   = loff2pgoff(offset);
	/* Mask lowest  'page_cluster'  worth of the lowest bits */
	reada   = ulong2pgoff(pgoff2ulong(reada) & ((~(0UL)) << page_cluster));

	for (i = 1 << page_cluster; i > 0;
	     --i, reada = ulong2pgoff(pgoff2ulong(reada)+1))
		new_page = try_to_read_ahead(file, reada, new_page);

	if (!new_page)
		new_page = page_cache_alloc();
	if (!new_page)
		goto oom;

	/*
	 * During getting the above page we might have slept,
	 * so we need to re-check the situation with the page
	 * cache.. The page we just got may be useful if we
	 * can't share, so don't get rid of it here.
	 */
	page = find_page(inode, pgoff);
	if (page)
		goto found_page;

	/*
	 * Now, create a new page-cache page from the page we got
	 */
	page = page_cache_entry(new_page);
	new_page = 0;
	add_to_page_cache(page, inode, pgoff, hash);

	if (inode->i_op->readpage(file, page) != 0)
		goto failure;

	goto found_page;

page_locked_wait:
	__wait_on_page(page);
	if (PageUptodate(page))
		goto success;
	
page_read_error:
	/*
	 * Umm, take care of errors if the page isn't up-to-date.
	 * Try to re-read it _once_. We do this synchronously,
	 * because there really aren't any performance issues here
	 * and we need to check for errors.
	 */
	if (inode->i_op->readpage(file, page) != 0)
		goto failure;
	wait_on_page(page);
	if (PageError(page))
		goto failure;
	if (PageUptodate(page))
		goto success;

	/*
	 * Things didn't work out. Return zero to tell the
	 * mm layer so, possibly freeing the page cache page first.
	 */
failure:
	page_cache_release(page);
	if (new_page)
		page_cache_free(new_page);
no_page:
	return 0;

release_and_oom:
	page_cache_release(page);
oom:
	return -1;
}

/*
 * Tries to write a shared mapped page to its backing store. May return -EIO
 * if the disk is full.
 */
static inline int do_write_page(struct inode * inode, struct file * file,
				const char * page, loff_t offset)
{
	int retval;
	loff_t size;
	loff_t loff = offset;
	mm_segment_t old_fs;

	size = offset + PAGE_SIZE;
	/* refuse to extend file size.. */
	if (S_ISREG(inode->i_mode)) {
		if (size > inode->i_size)
			size = inode->i_size;
		/* Ho humm.. We should have tested for this earlier */
		if (size < offset)
			return -EIO;
	}
	size -= offset;
	old_fs = get_fs();
	set_fs(KERNEL_DS);
	retval = -EIO;
	if (size == file->f_op->write(file, (const char *) page, size, &loff))
		retval = 0;
	set_fs(old_fs);
	return retval;
}

static int filemap_write_page(struct vm_area_struct * vma,
			      loff_t offset,
			      unsigned long page)
{
	int result;
	struct file * file;
	struct dentry * dentry;
	struct inode * inode;

	file = vma->vm_file;
	dentry = file->f_dentry;
	inode = dentry->d_inode;
	if (!file->f_op->write)
		return -EIO;

	/*
	 * If a task terminates while we're swapping the page, the vma and
	 * and file could be released ... increment the count to be safe.
	 */
	file->f_count++;
	fs_down(&inode->i_sem);
	result = do_write_page(inode, file, (const char *) page, offset);
	fs_up(&inode->i_sem);
	fput(file);
	return result;
}


/*
 * The page cache takes care of races between somebody
 * trying to swap something out and swap something in
 * at the same time..
 */
int filemap_swapout(struct vm_area_struct * vma, struct page * page)
{
	return filemap_write_page(vma, pgoff2loff(page->index), page_address(page));
}

static inline int filemap_sync_pte(pte_t * ptep, struct vm_area_struct *vma,
	unsigned long address, unsigned int flags)
{
	pte_t pte = *ptep;
	unsigned long page;
	int error;

	if (!(flags & MS_INVALIDATE)) {
		if (!pte_present(pte))
			return 0;
		if (!pte_dirty(pte))
			return 0;
		flush_page_to_ram(pte_page(pte));
		flush_cache_page(vma, address);
		set_pte(ptep, pte_mkclean(pte));
		flush_tlb_page(vma, address);
		page = pte_page(pte);
		atomic_inc(&page_cache_entry(page)->count);
	} else {
		if (pte_none(pte))
			return 0;
		flush_cache_page(vma, address);
		pte_clear(ptep);
		flush_tlb_page(vma, address);
		if (!pte_present(pte)) {
			swap_free(pte_val(pte));
			return 0;
		}
		page = pte_page(pte);
		if (!pte_dirty(pte) || flags == MS_INVALIDATE) {
			page_cache_free(page);
			return 0;
		}
	}
	error = filemap_write_page(vma, address - vma->vm_start + vma->vm_offset, page);
	page_cache_free(page);
	return error;
}

static inline int filemap_sync_pte_range(pmd_t * pmd,
	unsigned long address, unsigned long size, 
	struct vm_area_struct *vma, unsigned long offset, unsigned int flags)
{
	pte_t * pte;
	unsigned long end;
	int error;

	if (pmd_none(*pmd))
		return 0;
	if (pmd_bad(*pmd)) {
		printk("filemap_sync_pte_range: bad pmd (%08lx)\n", pmd_val(*pmd));
		pmd_clear(pmd);
		return 0;
	}
	pte = pte_offset(pmd, address);
	offset += address & PMD_MASK;
	address &= ~PMD_MASK;
	end = address + size;
	if (end > PMD_SIZE)
		end = PMD_SIZE;
	error = 0;
	do {
		error |= filemap_sync_pte(pte, vma, address + offset, flags);
		address += PAGE_SIZE;
		pte++;
	} while (address < end);
	return error;
}

static inline int filemap_sync_pmd_range(pgd_t * pgd,
	unsigned long address, unsigned long size, 
	struct vm_area_struct *vma, unsigned int flags)
{
	pmd_t * pmd;
	unsigned long offset, end;
	int error;

	if (pgd_none(*pgd))
		return 0;
	if (pgd_bad(*pgd)) {
		printk("filemap_sync_pmd_range: bad pgd (%08lx)\n", pgd_val(*pgd));
		pgd_clear(pgd);
		return 0;
	}
	pmd = pmd_offset(pgd, address);
	offset = address & PGDIR_MASK;
	address &= ~PGDIR_MASK;
	end = address + size;
	if (end > PGDIR_SIZE)
		end = PGDIR_SIZE;
	error = 0;
	do {
		error |= filemap_sync_pte_range(pmd, address, end - address, vma, offset, flags);
		address = (address + PMD_SIZE) & PMD_MASK;
		pmd++;
	} while (address < end);
	return error;
}

static int filemap_sync(struct vm_area_struct * vma, unsigned long address,
	size_t size, unsigned int flags)
{
	pgd_t * dir;
	unsigned long end = address + size;
	int error = 0;

	dir = pgd_offset(vma->vm_mm, address);
	flush_cache_range(vma->vm_mm, end - size, end);
	while (address < end) {
		error |= filemap_sync_pmd_range(dir, address, end - address, vma, flags);
		address = (address + PGDIR_SIZE) & PGDIR_MASK;
		dir++;
	}
	flush_tlb_range(vma->vm_mm, end - size, end);
	return error;
}

/*
 * This handles (potentially partial) area unmaps..
 */
static void filemap_unmap(struct vm_area_struct *vma, unsigned long start, size_t len)
{
	filemap_sync(vma, start, len, MS_ASYNC);
}

/*
 * Shared mappings need to be able to do the right thing at
 * close/unmap/sync. They will also use the private file as
 * backing-store for swapping..
 */
static struct vm_operations_struct file_shared_mmap = {
	NULL,			/* no special open */
	NULL,			/* no special close */
	filemap_unmap,		/* unmap - we need to sync the pages */
	NULL,			/* no special protect */
	filemap_sync,		/* sync */
	NULL,			/* advise */
	filemap_nopage,		/* nopage */
	NULL,			/* wppage */
	filemap_swapout,	/* swapout */
	NULL,			/* swapin */
};

/*
 * Private mappings just need to be able to load in the map.
 *
 * (This is actually used for shared mappings as well, if we
 * know they can't ever get write permissions..)
 */
static struct vm_operations_struct file_private_mmap = {
	NULL,			/* open */
	NULL,			/* close */
	NULL,			/* unmap */
	NULL,			/* protect */
	NULL,			/* sync */
	NULL,			/* advise */
	filemap_nopage,		/* nopage */
	NULL,			/* wppage */
	NULL,			/* swapout */
	NULL,			/* swapin */
};

/* This is used for a general mmap of a disk file */

int generic_file_mmap(struct file * file, struct vm_area_struct * vma)
{
	struct vm_operations_struct * ops;
	struct inode *inode = file->f_dentry->d_inode;

	if ((vma->vm_flags & VM_SHARED) && (vma->vm_flags & VM_MAYWRITE)) {
		ops = &file_shared_mmap;
		/* share_page() can only guarantee proper page sharing if
		 * the offsets are all page aligned. */
		if (vma->vm_offset & (PAGE_SIZE - 1))
			return -EINVAL;
	} else {
		ops = &file_private_mmap;
		if (inode->i_op && inode->i_op->bmap &&
		    (vma->vm_offset & (inode->i_sb->s_blocksize - 1)))
			return -EINVAL;
	}
	if (!inode->i_sb || !S_ISREG(inode->i_mode))
		return -EACCES;
	if (!inode->i_op || !inode->i_op->readpage)
		return -ENOEXEC;
	UPDATE_ATIME(inode);
	vma->vm_ops = ops;
	return 0;
}


/*
 * The msync() system call.
 */

static int msync_interval(struct vm_area_struct * vma,
	unsigned long start, unsigned long end, int flags)
{
	if (vma->vm_file && vma->vm_ops && vma->vm_ops->sync) {
		int error;
		error = vma->vm_ops->sync(vma, start, end-start, flags);
		if (!error && (flags & MS_SYNC)) {
			struct file * file = vma->vm_file;
			if (file) {
				struct dentry * dentry = file->f_dentry;
				struct inode * inode = dentry->d_inode;
				fs_down(&inode->i_sem);
				error = file_fsync(file, dentry);
				fs_up(&inode->i_sem);
			}
		}
		return error;
	}
	return 0;
}

asmlinkage int sys_msync(unsigned long start, size_t len, int flags)
{
	unsigned long end;
	struct vm_area_struct * vma;
	int unmapped_error, error = -EINVAL;

	down(&current->mm->mmap_sem);
	lock_kernel();
	if (start & ~PAGE_MASK)
		goto out;
	len = (len + ~PAGE_MASK) & PAGE_MASK;
	end = start + len;
	if (end < start)
		goto out;
	if (flags & ~(MS_ASYNC | MS_INVALIDATE | MS_SYNC))
		goto out;
	error = 0;
	if (end == start)
		goto out;
	/*
	 * If the interval [start,end) covers some unmapped address ranges,
	 * just ignore them, but return -EFAULT at the end.
	 */
	vma = find_vma(current->mm, start);
	unmapped_error = 0;
	for (;;) {
		/* Still start < end. */
		error = -EFAULT;
		if (!vma)
			goto out;
		/* Here start < vma->vm_end. */
		if (start < vma->vm_start) {
			unmapped_error = -EFAULT;
			start = vma->vm_start;
		}
		/* Here vma->vm_start <= start < vma->vm_end. */
		if (end <= vma->vm_end) {
			if (start < end) {
				error = msync_interval(vma, start, end, flags);
				if (error)
					goto out;
			}
			error = unmapped_error;
			goto out;
		}
		/* Here vma->vm_start <= start < vma->vm_end < end. */
		error = msync_interval(vma, start, vma->vm_end, flags);
		if (error)
			goto out;
		start = vma->vm_end;
		vma = vma->vm_next;
	}
out:
	unlock_kernel();
	up(&current->mm->mmap_sem);
	return error;
}

static inline
struct page *__read_cache_page(struct inode *inode, 
			       unsigned long offset,
			       int (*filler)(void *,struct page*),
			       void *data)
{
	struct page **hash = page_hash(inode, offset);
	struct page *page;
	unsigned long cached_page = 0;
	int err;

	offset &= PAGE_CACHE_MASK;
repeat:
	page = __find_page(inode, offset, *hash);
	if (!page) {
		if (!cached_page) {
			cached_page = page_cache_alloc();
			if (!cached_page)
				return ERR_PTR(-ENOMEM);
			goto repeat;
		}
		page = page_cache_entry(cached_page);
		cached_page = 0;
		add_to_page_cache(page, inode, offset, hash);
		set_bit(PG_locked, &page->flags);
		err = filler(data, page);
		if (err < 0) {
			page_cache_release(page);
			page = ERR_PTR(err);
		}
	}
	if (cached_page)
		page_cache_free(cached_page);
	return page;
}

/*
 * Read into the page cache. If a page already exists,
 * and Page_Uptodate() is not set, try to fill the page.
 */
struct page *read_cache_page(struct inode *inode,
				unsigned long offset,
				int (*filler)(void *,struct page*),
				void *data)
{
	struct page *page = __read_cache_page(inode, offset, filler, data);
	int err;

	if (IS_ERR(page) || PageUptodate(page))
		goto out;

	wait_on_page(page);
	if (PageUptodate(page))
		goto out;

	set_bit(PG_locked, &page->flags);
	err = filler(data, page);
	if (err < 0) {
		page_cache_release(page);
		page = ERR_PTR(err);
	}
 out:
	return page;
}

/*
 * Write to a file through the page cache. This is mainly for the
 * benefit of NFS and possibly other network-based file systems.
 *
 * We currently put everything into the page cache prior to writing it.
 * This is not a problem when writing full pages. With partial pages,
 * however, we first have to read the data into the cache, then
 * dirty the page, and finally schedule it for writing. Alternatively, we
 * could write-through just the portion of data that would go into that
 * page, but that would kill performance for applications that write data
 * line by line, and it's prone to race conditions.
 *
 * Note that this routine doesn't try to keep track of dirty pages. Each
 * file system has to do this all by itself, unfortunately.
 *							okir@monad.swb.de
 */
ssize_t
generic_file_write(struct file *file, const char *buf,
		   size_t count, loff_t *ppos)
{
	struct dentry	*dentry = file->f_dentry; 
	struct inode	*inode = dentry->d_inode; 
	loff_t		pos = *ppos;
	unsigned long	limit = current->rlim[RLIMIT_FSIZE].rlim_cur;
	struct page	*page, **hash;
	unsigned long	page_cache = 0;
	unsigned long	written;
	long		status, sync;

	if (!inode->i_op || !inode->i_op->updatepage)
		return -EIO;

	if (file->f_error) {
		int error = file->f_error;
		file->f_error = 0;
		return error;
	}

	if (((ssize_t) count) < 0)
		return -EINVAL;

	sync    = file->f_flags & O_SYNC;
	written = 0;

	if (file->f_flags & O_APPEND)
		pos = inode->i_size;

	/*
	 * Check whether we've reached the file size limit.
	 */
	status = -EFBIG;
	if (limit != RLIM_INFINITY && pos >= limit) {
		send_sig(SIGXFSZ, current, 0);
		goto out;
	}

	/* L-F-S */
	if (!(file->f_flags & O_LARGEFILE) &&
	    S_ISREG(inode->i_mode) && count) {
		if (pos >= 0x7fffffff) /* pos@2G forbidden */
			goto out;

		if (pos + count > 0x7fffffff)
			count = 0x7fffffff - pos;
	}

	status  = 0;
	/*
	 * Check whether to truncate the write,
	 * and send the signal if we do.
	 */
	if (limit != RLIM_INFINITY && count > limit - pos)
		count = limit - pos;

	while (count) {
		unsigned long bytes, offset;
		pgoff_t pgpos = loff2pgoff(pos);
		char * dest;

		/*
		 * Try to find the page in the cache. If it isn't there,
		 * allocate a free page.
		 */
		offset = ((unsigned long)pos & ~PAGE_CACHE_MASK);
		bytes = PAGE_CACHE_SIZE - offset;
		if (bytes > count)
			bytes = count;

		hash = page_hash(inode, pgpos);
		page = __find_page(inode, pgpos, *hash);
		if (!page) {
			if (!page_cache) {
				page_cache = page_cache_alloc();
				if (page_cache)
					continue;
				status = -ENOMEM;
				break;
			}
			page = page_cache_entry(page_cache);
			add_to_page_cache(page, inode, pgpos, hash);
			page_cache = 0;
		}

		/* Get exclusive IO access to the page.. */
		wait_on_page(page);
		set_bit(PG_locked, &page->flags);

		if (inode->i_op->prepare_write)
			status = inode->i_op->prepare_write(file, page, offset, bytes);
		if (status < 0)
			goto unlock;

		/*
		 * Do the real work.. If the writer ends up delaying the write,
		 * the writer needs to increment the page use counts until he
		 * is done with the page.
		 */
		dest = (char *) page_address(page) + offset;
		if (dest != buf) { /* See comment in update_vm_cache_cond. */
			bytes -= copy_from_user(dest, buf, bytes);
			flush_dcache_page(page_address(page));
		}
		status = -EFAULT;
		if (bytes)
			status = inode->i_op->updatepage(file, page, offset, bytes, sync);

 unlock:
		/* Mark it unlocked again and drop the page.. */
		clear_bit(PG_locked, &page->flags);
		wake_up(&page->wait);
		page_cache_release(page);

		if (status < 0)
			break;

		written += status;
		count -= status;
		pos += status;
		buf += status;
	}
	*ppos = pos;
	if (pos > inode->i_size)
		inode->i_size = pos;

	if (page_cache)
		page_cache_free(page_cache);
out:
	return written ? written : status;
}

/*
 * Support routines for directory cacheing using the page cache.
 */

/*
 * Finds the page at the specified offset, installing a new page
 * if requested.  The count is incremented and the page is locked.
 *
 * Note: we don't have to worry about races here, as the caller
 * is holding the inode semaphore.
 */
unsigned long get_cached_page(struct inode * inode, pgoff_t pgoff, int new)
{
	struct page * page;
	struct page ** hash;
	unsigned long page_cache = 0;

	hash = page_hash(inode, pgoff);
	page = __find_page(inode, pgoff, *hash);
	if (!page) {
		if (!new)
			goto out;
		page_cache = page_cache_alloc();
		if (!page_cache)
			goto out;
		clear_page(page_cache);
		page = page_cache_entry(page_cache);
		add_to_page_cache(page, inode, pgoff, hash);
	}
	if (atomic_read(&page->count) != 2)
		printk(KERN_ERR "get_cached_page: page count=%d\n",
			atomic_read(&page->count));
	if (test_bit(PG_locked, &page->flags))
		printk(KERN_ERR "get_cached_page: page already locked!\n");
	set_bit(PG_locked, &page->flags);
	page_cache = page_address(page);

out:
	return page_cache;
}

/*
 * Unlock and free a page.
 */
void put_cached_page(unsigned long addr)
{
	struct page * page = page_cache_entry(addr);

	if (!test_bit(PG_locked, &page->flags))
		printk("put_cached_page: page not locked!\n");
	if (atomic_read(&page->count) != 2)
		printk("put_cached_page: page count=%d\n", 
			atomic_read(&page->count));
	clear_bit(PG_locked, &page->flags);
	wake_up(&page->wait);
	page_cache_release(page);
}

void __init page_cache_init(unsigned long memory_size)
{
	unsigned long htable_size;
	long order;

	htable_size  = memory_size >> PAGE_SHIFT;
	htable_size *= sizeof(struct page *);
	for(order = 0; (PAGE_SIZE << order) < htable_size; order++)
		;

	do {
		unsigned long tmp = (PAGE_SIZE << order) / sizeof(struct page *);

		page_hash_mask = (tmp - 1UL);

		page_hash_bits = 0;
		while((tmp >>= 1UL) != 0UL)
			page_hash_bits++;

		page_hash_table = (struct page **)
			__get_free_pages(GFP_ATOMIC, order);
	} while(page_hash_table == NULL && --order >= 0L);

	printk("Page cache hash table entries: %d (order %ld, %ldk)\n",
	       (1 << page_hash_bits), order, (1UL << order) * PAGE_SIZE / 1024);
	if (!page_hash_table)
		panic("Failed to allocate page hash table\n");
	memset(page_hash_table, 0,
	       (PAGE_HASH_MASK + 1UL) * sizeof(struct page *));
}
